import json
from typing import Any

from asyncblink import signal  # type: ignore
from fastapi import APIRouter, Depends, WebSocket

from ..d1_response import D1Encoder
from ..deps import read_access_websocket
from ..models.environment_info import EnvironmentInfoWrapper, Version
from ..models.server_stats import ServerStatsListWrapper, ServerStatsWrapper
from ..settings import settings
from ..signals import (
    APP_LOG,
    NOTEBOOKS_UPDATED,
    RCLONE_UPDATE,
    SERVER_STATS_UPDATE,
    WATCHDOG_FILE_SYSTEM_EVENT,
)
from ..storage import environment_dump, stats_manager
from ..utils import websocket_poll

router = APIRouter()


@router.get("/version", response_model=Version)
def get_server_version() -> Any:
    """
    Get D1 server version number
    """
    return Version(version=settings.VERSION)


@router.get("/stats", response_model=ServerStatsListWrapper)
def get_server_stats(limit: int = 100, skip: int = 0) -> Any:
    """
    Get server stats about memory, cpu, disk, and gpu
    """
    stats = stats_manager.get_stats(skip, skip + limit)
    return ServerStatsListWrapper(server_stats=stats)


@router.get("/latest_stats", response_model=ServerStatsWrapper)
def get_latest_server_stats(limit: int = 100, skip: int = 0) -> Any:
    """
    Get latest server stats about memory, cpu, disk, and gpu
    """
    latest_stats = stats_manager.get_latest()
    return ServerStatsWrapper(server_stats=latest_stats)


@router.websocket("/ws")
async def server_stats_websocket(
    websocket: WebSocket, read_access: bool = Depends(read_access_websocket)
):
    await websocket.accept()
    msgs = []

    async def receive_notebooks_updated(sender, notebooks, **kwargs):
        nb_dicts = [
            {"uuid": nb_uuid, "name": nb.metadata.jupyter_d1.name}
            for (nb_uuid, nb) in notebooks.items()
        ]
        wrapper_dict = {"notebooks": nb_dicts}
        json_string = json.dumps(wrapper_dict, cls=D1Encoder)
        msgs.append(json_string)

    signal(NOTEBOOKS_UPDATED).connect(receive_notebooks_updated)

    async def receive_rclone_update(sender, stats, **kwargs):
        wrapper_dict = {"rclone_stats": stats}
        json_string = json.dumps(wrapper_dict, cls=D1Encoder)
        msgs.append(json_string)

    signal(RCLONE_UPDATE).connect(receive_rclone_update)

    async def receive_watchdog_event(
        sender, event_type, src_path, dest_path, **kwargs
    ):
        wrapper_dict = {
            "watchdog_event": {
                "event_type": event_type,
                "src_path": src_path,
                "dest_path": dest_path,
            }
        }
        json_string = json.dumps(wrapper_dict, cls=D1Encoder)
        msgs.append(json_string)

    signal(WATCHDOG_FILE_SYSTEM_EVENT).connect(receive_watchdog_event)

    async def receive_server_stats(sender, server_stats, **kwargs):
        wrapper_dict = {"server_stats": server_stats.dict()}
        json_string = json.dumps(wrapper_dict, cls=D1Encoder)
        msgs.append(json_string)

    signal(SERVER_STATS_UPDATE).connect(receive_server_stats)

    async def receive_app_log(sender, msg, **kwargs):
        wrapper_dict = {"log": msg}
        json_string = json.dumps(wrapper_dict, cls=D1Encoder)
        msgs.append(json_string)

    signal(APP_LOG).connect(receive_app_log)

    await websocket_poll(websocket, msgs)


@router.get("/environment", response_model=EnvironmentInfoWrapper)
def get_environment() -> Any:
    """
    Get environment info
    """
    environment_info = environment_dump.dump_environment()
    return EnvironmentInfoWrapper(environment_info=environment_info)



from typing import Dict, List

from pydantic import BaseModel

from .base_wrapper import BaseWrapper
from .JSONType import JSONType


class KernelSpecDetail(BaseModel):
    display_name: str
    language: str
    interrupt_mode: str
    metadata: JSONType
    argv: List[str]
    env: Dict[str, str]


class KernelSpec(BaseModel):
    resource_dir: str
    spec: KernelSpecDetail
    kernel_name: str


class KernelSpecsWrapper(BaseWrapper):
    kernel_specs: List[KernelSpec]



from datetime import datetime
from typing import List
from uuid import UUID

from pydantic import BaseModel


class JupyterD1Metadata(BaseModel):
    cell_uuid: str


class KernelPayload(BaseModel):
    pass


class KernelHeader(BaseModel):
    # date: DateTime
    msg_id: str
    msg_type: str
    session: str
    username: str
    version: str


class KernelMetadata(BaseModel):
    dependencies_met: bool
    engine: str
    jupyter_d1: JupyterD1Metadata
    # started: DateTime
    status: str


class KernelContent(BaseModel):
    execution_count: int
    payload: List[KernelPayload]
    status: str
    # user_expression: str


class KernelMessage(BaseModel):
    msg_id: str
    msg_type: str
    channel: str
    header: KernelHeader
    parent_header: KernelHeader
    metadata: KernelMetadata
    content: KernelContent
    # buffers: [KernelBuffer]


class KernelIsAlive(BaseModel):
    kernel_id: UUID
    kernel_is_alive: bool


class KernelsIdle(BaseModel):
    idle: bool
    last_idle: datetime



from jupyter_d1.settings import settings


def test_settings():
    assert settings.SECRET_KEY == "jukugoeir32*"
    assert settings.PUSH_NOTE_SECRET_KEY == "12iuf49f))("
    assert settings.WORK_NODE_ID == 433451
    assert settings.MOTHERSHIP_URL == "https://staging.callistoapp.com/api/v1"
    assert settings.WATCHDOG_ENABLED



import os
from pathlib import Path
from typing import Optional
from uuid import UUID, uuid4

from fastapi.logger import logger
from nbformat.notebooknode import NotebookNode  # type: ignore


class NotebookValidator:
    def _validate_jupyter_metadata_node(self, node):
        """
        Create the jupyter_d1 section in the node's metadata if it
        doesn't exist.
        """
        if "jupyter_d1" not in node.metadata.keys():
            node.metadata.jupyter_d1 = NotebookNode()

    def _validate_jupyter_metadata(self, notebook_node: NotebookNode):
        "Ensure all the notebook nodes have jupyter_d1 section"
        self._validate_jupyter_metadata_node(notebook_node)
        for cell in notebook_node.cells:
            self._validate_jupyter_metadata_node(cell)

    def _validate_uuid_node(
        self, node, force_new: bool = False, reset_state: bool = False
    ) -> UUID:
        "Return the uuid from a node's metadata.  Create one if it is missing."
        if "uuid" not in node.metadata.jupyter_d1.keys() or force_new:
            node.metadata.jupyter_d1.uuid = uuid4()
        if reset_state:
            node.metadata.jupyter_d1.execution_state = "idle"
            if "execution_count" in node:
                node.execution_count = None
        return node.metadata.jupyter_d1.uuid

    def _validate_uuids(
        self,
        notebook_node: NotebookNode,
        new_notebook_uuid: bool = False,
        reset_state: bool = False,
    ) -> UUID:
        "Ensure all the notebook nodes have uuids & return the notebook uuid."
        nb_uuid = self._validate_uuid_node(notebook_node, new_notebook_uuid)
        uuids = [nb_uuid]
        for cell in notebook_node.cells:
            cell_uuid = self._validate_uuid_node(cell, reset_state=reset_state)
            cell.metadata.jupyter_d1.notebook_uuid = nb_uuid
            uuids.append(cell_uuid)

        # Check for uniqueness in UUIDs
        uuid_set = set(uuids)
        if len(uuids) != len(uuid_set):
            for uuid in uuid_set:
                uuids.remove(uuid)
            logger.error(f"Duplicate UUID(s) found in validation. {uuids}")
            raise IndexError(f"Duplicate uuids in notebook: {uuids}")
        return nb_uuid

    def _validate_positions(self, notebook_node: NotebookNode):
        "Renumber all the cells according to their current position"
        for (idx, cell) in enumerate(notebook_node.cells):
            cell.metadata.jupyter_d1.position = idx

    def _validate_path(
        self,
        notebook_node: NotebookNode,
        directory: Optional[str],
        filename: str,
    ):
        "Set name and path attribute in metadata"
        notebook_node.metadata.jupyter_d1.name = filename
        path = f"{filename}.ipynb"
        if directory is not None:
            try:
                directory = str(Path(directory).resolve())
            except Exception as e:
                logger.error(e)
            path = f"{directory}/{path}"
        notebook_node.metadata.jupyter_d1.path = path

    def _validate_working_dir(
        self,
        notebook_node: NotebookNode,
        working_directory: Optional[str],
        directory: Optional[str],
    ):
        """
        Set working_directory attribute in metadata.

        Priority is working_directory parameter, existing metadata
        working_directory, and then defaults to directory
        """
        d1_meta = notebook_node.metadata.jupyter_d1
        if working_directory is None:
            working_directory = d1_meta.get("working_directory", None)
        if working_directory is None or not os.path.exists(working_directory):
            working_directory = directory
        if working_directory is not None:
            try:
                working_directory = str(Path(working_directory).resolve())
            except Exception as e:
                logger.error(e)
        d1_meta.working_directory = working_directory

    def validate(
        self,
        notebook_node: NotebookNode,
        directory: Optional[str] = None,
        filename: Optional[str] = None,
        working_directory: Optional[str] = None,
        new_notebook_uuid: bool = False,
        reset_state: bool = False,
    ):
        """
        Ensure all the notebook nodes have positions and uuids
        """
        self._validate_jupyter_metadata(notebook_node)
        self._validate_positions(notebook_node)
        self._validate_uuids(
            notebook_node, new_notebook_uuid, reset_state=reset_state
        )
        if filename is not None:
            self._validate_path(notebook_node, directory, filename)
        if directory is not None or working_directory is not None:
            self._validate_working_dir(
                notebook_node, working_directory, directory
            )



from jupyter_d1.undo_manager import UndoManager

logging = False


class Seq:
    def __init__(self, seq, undo_manager):
        "Wraps a list of int"
        self._seq = seq
        self.undo_manager = undo_manager

    def add(self, value=1):
        "Add 'value' to the end of the list."
        self._seq.append(value)
        self.undo_manager.add_action(self.remove, name=f"Add {value}")

    def remove(self):
        "Removes the last item from the list"
        value = self._seq.pop()
        kwargs = {"value": value}
        self.undo_manager.add_action(
            self.add, kwargs=kwargs, name=f"Remove {value}"
        )
        return value

    def bulk_add(self, values=[]):
        "Add values to the end of the list"
        count = len(values)
        self.undo_manager.pause()
        for value in values:
            self.add(value)
        self.undo_manager.unpause()

        kwargs = {"count": count}
        self.undo_manager.add_action(
            self.bulk_remove, kwargs=kwargs, name=f"Bulk add {len(values)}"
        )

    def bulk_remove(self, count):
        "Remove 'count' values from the list"
        self.undo_manager.pause()
        values = []
        for i in range(count):
            value = self.remove()
            values.append(value)
        self.undo_manager.unpause()

        values.reverse()  # for undo, need reverse order
        kwargs = {"values": values}
        self.undo_manager.add_action(
            self.bulk_add, kwargs=kwargs, name=f"Bulk remove {len(values)}"
        )

    @property
    def seq(self):
        return self._seq


class TestUndoManager:
    def test_undo_bulk(self):
        um = UndoManager(logging=logging)
        assert um.undo_count == 0
        assert um.redo_count == 0

        seq = Seq([11, 12], um)
        assert seq.seq == [11, 12]

        seq.bulk_add([25, 26, 27])
        assert seq.seq == [11, 12, 25, 26, 27]
        assert um.undo_count == 1
        assert um.redo_count == 0
        assert um.next_undo_name == "Bulk add 3"

        um.undo()
        assert seq.seq == [11, 12]
        assert um.undo_count == 0
        assert um.redo_count == 1
        assert um.next_redo_name == "Bulk add 3"

        um.redo()
        assert seq.seq == [11, 12, 25, 26, 27]
        assert um.undo_count == 1
        assert um.redo_count == 0
        assert um.next_undo_name == "Bulk add 3"

        seq.bulk_remove(2)
        assert seq.seq == [11, 12, 25]
        assert um.undo_count == 2
        assert um.redo_count == 0
        assert um.next_undo_name == "Bulk remove 2"

        um.undo()
        assert seq.seq == [11, 12, 25, 26, 27]
        assert um.undo_count == 1
        assert um.redo_count == 1
        assert um.next_redo_name == "Bulk remove 2"

        um.redo()
        assert seq.seq == [11, 12, 25]
        assert um.undo_count == 2
        assert um.redo_count == 0
        assert um.next_undo_name == "Bulk remove 2"

    def test_undo(self):
        um = UndoManager(logging=logging)
        assert um.undo_count == 0
        assert um.redo_count == 0

        seq = Seq([1, 2], um)
        assert seq.seq == [1, 2]

        seq.add(3)
        assert seq.seq == [1, 2, 3]
        assert um.undo_count == 1
        assert um.next_undo_name == "Add 3"
        assert um.redo_count == 0

        um.undo()
        assert seq.seq == [1, 2]
        assert um.undo_count == 0
        assert um.redo_count == 1
        assert um.next_redo_name == "Add 3"

        # extra undo() calls shouldn't do anything
        um.undo()
        um.undo()
        um.undo()
        assert seq.seq == [1, 2]
        assert um.undo_count == 0
        assert um.redo_count == 1
        assert um.next_redo_name == "Add 3"

        um.redo()
        assert seq.seq == [1, 2, 3]
        assert um.undo_count == 1
        assert um.next_undo_name == "Add 3"
        assert um.redo_count == 0

        # more redo() calls shouldn't do anything.
        um.redo()
        um.redo()
        um.redo()
        assert seq.seq == [1, 2, 3]
        assert um.undo_count == 1
        assert um.next_undo_name == "Add 3"
        assert um.redo_count == 0

        seq.remove()
        assert seq.seq == [1, 2]
        assert um.undo_count == 2
        assert um.next_undo_name == "Remove 3"
        assert um.redo_count == 0

        um.undo()
        assert seq.seq == [1, 2, 3]
        assert um.undo_count == 1
        assert um.next_undo_name == "Add 3"
        assert um.redo_count == 1
        assert um.next_redo_name == "Remove 3"

        um.redo()
        assert seq.seq == [1, 2]
        assert um.undo_count == 2
        assert um.next_undo_name == "Remove 3"
        assert um.redo_count == 0



from typing import Any, Dict

# This is a placeholder for JSONType until it gets officially support.
# 5/2020 -- waiting on mypy support for recursion.
# https://github.com/python/typing/issues/182
JSONType = Dict[str, Any]



import json
import os
import platform
import secrets
from pathlib import Path
from typing import Any, Dict, List, Optional

from pydantic import BaseSettings, validator

callisto_env = os.getenv("CALLISTO_ENV", "dev").lower()
dotenv_file = f"{callisto_env}.env"


class Settings(BaseSettings):
    pretty_name: str = "Jupyter D1"
    VERSION: int = 1
    OAUTH_TOKEN_URL: str = "/login/access-token"
    META_INFO_FILE: Optional[str] = None
    META_INFO: Dict[str, Any] = {}

    @validator("META_INFO", pre=True)
    def get_meta_info(
        cls, v: Dict[str, Any], values: Dict[str, Any]
    ) -> Dict[str, Any]:
        if len(v.keys()) > 0:
            return v
        meta_info_file = values.get("META_INFO_FILE", None)
        if meta_info_file is not None:
            with open(meta_info_file, "r") as f:
                return json.loads(f.read().rstrip())
        else:
            return {}

    WORK_NODE_ID: Optional[int] = None
    SECRET_KEY: str = ""
    PUSH_NOTE_SECRET_KEY: Optional[str] = None
    MOTHERSHIP_URL: Optional[str] = None

    @validator("WORK_NODE_ID", pre=True)
    def get_work_node_id(
        cls, v: Optional[int], values: Dict[str, Any]
    ) -> Optional[int]:
        meta_info = values.get("META_INFO", None)
        if meta_info is not None and "work_node_id" in meta_info:
            return meta_info["work_node_id"]
        return None

    @validator("SECRET_KEY", pre=True)
    def get_secret_key(cls, v: str, values: Dict[str, Any]) -> str:
        if len(v) > 0:
            return v
        if callisto_env in ("dev", "local"):
            return secrets.token_urlsafe(32)
        meta_info = values.get("META_INFO", None)
        if meta_info is not None and "auth_secret" in meta_info:
            return meta_info["auth_secret"]
        raise ValueError(v)

    @validator("PUSH_NOTE_SECRET_KEY", pre=True)
    def get_push_note_secret_key(
        cls, v: Optional[str], values: Dict[str, Any]
    ) -> Optional[str]:
        if v is not None and len(v) > 0:
            return v
        meta_info = values.get("META_INFO", None)
        if meta_info is not None and "push_note_secret" in meta_info:
            return meta_info["push_note_secret"]
        return None

    @validator("MOTHERSHIP_URL", pre=True)
    def get_mothership_url(
        cls, v: Optional[str], values: Dict[str, Any]
    ) -> Optional[str]:
        meta_info = values.get("META_INFO", None)
        if meta_info is not None and "env" in meta_info:
            if meta_info["env"] == "prod":
                return "https://app.callistoapp.com/api/v1"
            else:
                return "https://staging.callistoapp.com/api/v1"
        return None

    # Login enabled for local environment
    LOGIN_ENABLED: bool = callisto_env == "dev"
    LOGIN_TOKEN: Optional[str] = None
    # 60 minutes * 24 hours * 8 days = 8 days
    ACCESS_TOKEN_EXPIRE_MINUTES: int = 60 * 24 * 8

    # This is the root directory exposed via WebDAV
    ROOT_DIR: str = os.getcwd()

    @validator("ROOT_DIR")
    def absolute_root_dir(cls, v: str, values: Dict[str, Any]) -> str:
        return str(Path(v).resolve())

    # Rclone settings
    RCLONE_STATS_POLL = True
    RCLONE_STATS_POLLING_INTERVAL = 20  # seconds

    # Server stats
    SERVER_STATS_POLLING_INTERVAL = 5  # seconds
    SERVER_STATS_TTL = 24  # hours

    DAV_SETTINGS: Dict[str, Any] = {}

    @validator("DAV_SETTINGS", pre=True)
    def get_dav_settings(
        cls, v: Dict[str, Any], values: Dict[str, Any]
    ) -> Dict[str, Any]:
        root_dir = values.get("ROOT_DIR", None)
        if root_dir is None:
            raise ValueError(v)
        Path(root_dir).mkdir(parents=True, exist_ok=True)

        return {
            "host": "0.0.0.0",
            "port": 8080,
            "mount_path": "/dav",
            "provider_mapping": {root_dir.lower(): root_dir},
            "verbose": 3,
            "http_authenticator": {
                "domain_controller": "jupyter_d1.dav_auth.JWTDomainController",
                "accept_basic": True,
                "accept_digest": False,
                "default_to_digest": False,
            },
            # Taken from default config
            # (https://wsgidav.readthedocs.io/en/latest/user_guide_configure.html)  # noqa
            "dir_browser": {
                # Render HTML listing for GET requests on collections
                "enable": True,
                # List of fnmatch patterns:
                "ignore": [
                    ".DS_Store",  # macOS folder meta data
                    "._*",  # macOS hidden data files
                    "Thumbs.db",  # Windows image previews
                ],
                "icon": True,
                # Raw HTML code, appended as footer (True: use a default)
                "response_trailer": True,
                "show_user": True,  # Show authenticated user an realm
                # Send <dm:mount> response if request URL contains '?davmount'
                "davmount": False,
                # Invoke MS Offce documents for editing using WebDAV
                "ms_sharepoint_support": True,
                # The path to the directory that contains template.html and
                # associated assets.
                # The default is the htdocs directory within the dir_browser
                # directory.
                "htdocs_path": None,
            },
        }

    WATCHDOG_IGNORE_PATTERNS: List[str] = [".DS_Store", "._*"]
    WATCHDOG_IGNORE_DIRS: List[str] = ["data"]
    WATCHDOG_ENABLED: bool = False

    @validator("WATCHDOG_ENABLED")
    def get_watchdog_enabled(cls, v: bool, values: Dict[str, Any]) -> bool:
        if v is True:
            return True
        meta_info = values.get("META_INFO", None)
        if meta_info is not None and "env" in meta_info:
            if meta_info["env"] == "test":
                return True
        if platform.system().startswith("Darwin"):
            return False
        return True

    # Number of elements (approximately) at which to abbreviate variables
    # in the variable explorer
    VAR_ABBREV_LEN: int = 50

    LOAD_CALLISTOR_FROM_GITHUB: bool = True
    CALLISTOR_GITHUB_BRANCH: str = "dev"

    @validator("CALLISTOR_GITHUB_BRANCH", pre=True)
    def get_callistor_branch(
        cls, v: Optional[str], values: Dict[str, Any]
    ) -> Optional[str]:
        meta_info = values.get("META_INFO", None)
        if meta_info is not None and "env" in meta_info:
            if meta_info["env"] == "prod":
                return "master"
        return "dev"

    class Config:
        case_sensitive = True
        env_prefix = "CALLISTO_"
        env_file = dotenv_file


settings = Settings()



from typing import Optional

from pydantic import BaseModel

from .base_wrapper import BaseWrapper


class Permission(BaseModel):
    id: int
    user_id: int
    work_node_id: int
    read_access: Optional[bool]
    write_access: Optional[bool]


class PermissionWrapper(BaseWrapper):
    permission: Permission



from ...utils import NotebookNode
from ..workdir_manager import WorkDirManager

raw_chdir_code = 'setwd("{directory}")'
raw_cwd_code = "noquote(getwd())"


class RWorkDirManager(WorkDirManager):
    def get_chdir_code(self, directory: str) -> str:
        return f"{raw_chdir_code.format(directory=directory)}\n{raw_cwd_code}"

    def get_cwd_code(self) -> str:
        return raw_cwd_code

    def extract_cwd(self, response: NotebookNode) -> str:
        return response["data"]["text/plain"].split(" ", 1)[1]

    def parse_chdir_response(self, response: NotebookNode) -> str:
        return self.extract_cwd(response)

    def parse_cwd_response(self, response: NotebookNode) -> str:
        return self.extract_cwd(response)



from ..settings import settings
from .environment_dump import EnvironmentDump
from .kernel_logger import KernelLogger
from .kernel_manager import KernelManager
from .notebook_manager import NBMException  # noqa
from .notebook_manager import NotebookManager
from .stats_manager import StatsManager

kmanager = KernelManager()
manager = NotebookManager(kmanager)
stats_manager = StatsManager(
    settings.SERVER_STATS_POLLING_INTERVAL * settings.SERVER_STATS_TTL * 60
)
environment_dump = EnvironmentDump()

# Instantiate a logger to echo out the messages from the kernels
klogger = KernelLogger()



from pydantic import BaseModel

from .base_wrapper import BaseWrapper


class Token(BaseModel):
    access_token: str
    token_type: str


class TokenPayload(BaseModel):
    sub: str


class TokenWrapper(BaseWrapper):
    token: Token



import asyncio
import json
import os
import pathlib
from asyncio.exceptions import TimeoutError
from datetime import timedelta
from typing import Any, Dict, List

from fastapi import WebSocket
from fastapi.encoders import jsonable_encoder
from fastapi.testclient import TestClient

from jupyter_d1 import security
from jupyter_d1.models.permission import Permission
from jupyter_d1.settings import settings

WEBSOCKET_RECEIVE_TIMEOUT = float(
    os.environ.get("TEST_WEBSOCKET_RECEIVE_TIMEOUT", 4.0)
)

# Another directory to use during testing, cleared by a test fixture after
# every test
workdir_1 = str(pathlib.Path("/tmp/test_nb_2").resolve())


def get_superuser_token_headers(client: TestClient) -> Dict[str, str]:
    r = client.get(
        "/login/access-token", headers={"Authorization": "test9token_4"}
    )
    tokens = r.json()["token"]
    a_token = tokens["access_token"]
    headers = {"Authorization": f"Bearer {a_token}"}
    return headers


def create_auth_headers(token: str) -> Dict[str, str]:
    headers = {"Authorization": f"Bearer {token}"}
    return headers


def create_auth_token(permission: Permission) -> str:
    access_token_expires = timedelta(
        minutes=settings.ACCESS_TOKEN_EXPIRE_MINUTES
    )
    a_token = security.create_access_token(
        json.dumps(jsonable_encoder(permission)),
        expires_delta=access_token_expires,
    )
    return a_token


def get_superuser_token() -> str:
    permission = Permission(
        id=-1, user_id=-1, work_node_id=-1, write_access=True, read_access=True
    )

    return create_auth_token(permission)


def get_read_only_token() -> str:
    permission = Permission(
        id=-1,
        user_id=-1,
        work_node_id=-1,
        write_access=False,
        read_access=True,
    )

    return create_auth_token(permission)


def get_permissionless_token() -> str:
    permission = Permission(
        id=-1,
        user_id=-1,
        work_node_id=-1,
        write_access=False,
        read_access=False,
    )

    return create_auth_token(permission)


def get_readonly_token_headers() -> Dict[str, str]:
    return create_auth_headers(get_read_only_token())


def get_permissionless_token_headers() -> Dict[str, str]:
    return create_auth_headers(get_permissionless_token())


async def receive_json(
    websocket: WebSocket,
    timeout: float = WEBSOCKET_RECEIVE_TIMEOUT,
    msg_types: List[str] = [],
    max_retries: int = 20,
) -> Dict[str, Any]:
    """ 
    timeout -- timeout value for waiting on websocket data
    msg_types -- list of messages types to look for in the websocket stream.
        Types not included in the list will be ignored.
        The default value of [] means no filtering -- return all messages
    max_retries -- the max number of messages that can be ignore (that don't 
        match the msg_types list).  Default is 20.
    """
    count = 0
    while True:
        data = await asyncio.wait_for(
            websocket.receive_json(), timeout=timeout
        )

        # early exit if there's no filtering to do
        if len(msg_types) == 0:
            return data

        keys = list(data.keys())
        for msg_type in msg_types:
            if msg_type in keys:
                return data
        assert count < max_retries, "Too many retries in receive_json"
        count += 1


async def wait_for_event(
    websocket: WebSocket,
    event_type: str,
    tries: int = 15,
    timeout: float = WEBSOCKET_RECEIVE_TIMEOUT,
) -> Dict[str, Any]:

    return await receive_json(
        websocket, timeout=timeout, msg_types=[event_type], max_retries=tries
    )


async def collect_websocket_messages(
    websocket: WebSocket, timeout: float = WEBSOCKET_RECEIVE_TIMEOUT
) -> List[Dict[str, Any]]:
    """ 
    Start listening to the websocket and collect all messages until the 
    timeout is exceeded waiting on the next message.  Return all received 
    messages.
    """

    msgs = []
    try:
        while True:
            msg = await receive_json(websocket, timeout)
            msgs.append(msg)
    except TimeoutError:
        pass

    return msgs


def filter_websocket_collection(
    messages: List[Dict[str, Any]], msg_type: str
) -> List[Dict[str, Any]]:
    """ 
    Give a list of websocket messages, strip out and return the messages that
    contain `msg_type` as a top level key.
    """

    def has_key(a_dict: Dict[str, Any]) -> bool:
        return msg_type in a_dict.keys()

    return list(filter(has_key, messages))



import json
from datetime import timedelta
from typing import Any, Optional

from fastapi import APIRouter, Depends, Header, HTTPException
from fastapi.encoders import jsonable_encoder

from .. import deps, security
from ..models.permission import Permission, PermissionWrapper
from ..models.token import TokenWrapper
from ..settings import settings

router = APIRouter()


@router.get("/login/access-token", response_model=TokenWrapper)
def login_access_token(authorization: Optional[str] = Header(None)) -> Any:
    """
    OAuth2 compatible token login, get an access token for future requests.
    """
    if settings.LOGIN_ENABLED is not True and settings.LOGIN_TOKEN is None:
        raise HTTPException(
            400, "Login not enabled, login with the Callisto mothership"
        )
    if (
        settings.LOGIN_TOKEN is not None
        and settings.LOGIN_TOKEN != authorization
    ):
        raise HTTPException(401, "Login token does not match")

    permission = Permission(
        id=-1, user_id=-1, work_node_id=-1, write_access=True, read_access=True
    )

    access_token_expires = timedelta(
        minutes=settings.ACCESS_TOKEN_EXPIRE_MINUTES
    )
    token = {
        "access_token": security.create_access_token(
            json.dumps(jsonable_encoder(permission)),
            expires_delta=access_token_expires,
        ),
        "token_type": "bearer",
    }
    return TokenWrapper(token=token)


@router.post("/login/test-token", response_model=PermissionWrapper)
def test_token(
    permission: Permission = Depends(deps.get_current_permission),
) -> Any:
    """
    Test access token
    """
    return PermissionWrapper(permission=permission)



from pytest_mock import MockFixture  # type: ignore

from jupyter_d1.commands import execute_d1_command


def test_notify(mocker: MockFixture):
    requests = mocker.patch("jupyter_d1.commands.requests")

    execute_d1_command(
        '{"command_type":"notify",' '"title":"atitle",' '"message":"test6e4"}'
    )
    requests.post.assert_called_with(
        f"https://staging.callistoapp.com/api/v1/work_nodes/433451/push",
        json={
            "secret": "12iuf49f))(",
            "title": "atitle",
            "message": "test6e4",
        },
        timeout=30,
    )


def test_notify_invalid(mocker: MockFixture):
    requests = mocker.patch("jupyter_d1.commands.requests")

    execute_d1_command(
        '{"command_type":"ju",' '"title":"atitle",' '"message":"test6e4"}'
    )
    requests.post.assert_not_called()



from ..kernel_definition import KernelDefinition
from .vars_manager import RVarsManager, VarsManager
from .workdir_manager import RWorkDirManager, WorkDirManager


class RKernelDefinition(KernelDefinition):
    def create_vars_manager(self) -> VarsManager:
        return RVarsManager()

    def create_workdir_manager(self, workdir: str) -> WorkDirManager:
        return RWorkDirManager(workdir)


kernel_definition = RKernelDefinition()



from abc import ABC, abstractmethod

from ..utils import NotebookNode


class WorkDirManager(ABC):
    def __init__(self, workdir: str):
        self._workdir = workdir

    @property
    def workdir(self) -> str:
        return self._workdir

    @abstractmethod
    def get_chdir_code(self, directory: str) -> str:
        raise NotImplementedError

    @abstractmethod
    def get_cwd_code(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def parse_chdir_response(self, response: NotebookNode) -> str:
        raise NotImplementedError

    @abstractmethod
    def parse_cwd_response(self, response: NotebookNode) -> str:
        raise NotImplementedError

    def parse_chdir_output(self, output: NotebookNode) -> str:
        self._workdir = self.parse_chdir_response(output)
        return self._workdir

    def parse_cwd_output(self, output: NotebookNode) -> str:
        self._workdir = self.parse_cwd_response(output)
        return self._workdir



from typing import Any, Dict, List, Optional

from pydantic import BaseModel

from .base_wrapper import BaseWrapper


class OSInfo(BaseModel):
    platform: str
    name: str
    raw_name: str
    uname: Dict[str, Any]


class PythonVersionInfo(BaseModel):
    major: int
    minor: int
    micro: int
    releaselevel: str
    serial: int


class PythonInfo(BaseModel):
    version: str
    executable: str
    pythonpath: List[str]
    version_info: PythonVersionInfo
    packages: Optional[Dict[str, str]]


class ProcessInfo(BaseModel):
    argv: List[str]
    cwd: str
    user: str
    pid: int
    environ: Dict[str, Any]


class CondaInfo(BaseModel):
    version: str


class EnvironmentInfo(BaseModel):
    os: OSInfo
    python: PythonInfo
    config: Dict[str, Any]
    process: ProcessInfo
    conda: Optional[CondaInfo]
    cores: int
    version: int


class EnvironmentInfoWrapper(BaseWrapper):
    environment_info: EnvironmentInfo


class Version(BaseModel):
    version: int



from typing import List

from .base_wrapper import BaseWrapper
from .JSONType import JSONType


class NotebookWrapper(BaseWrapper):
    notebook: JSONType


class NotebooksWrapper(BaseWrapper):
    notebooks: List[JSONType]


class NotebookPath(BaseWrapper):
    path: str



import asyncio

from asyncblink import signal  # type: ignore

from jupyter_d1.signals import (
    APP_SHUTDOWN,
    CONTROL_CHANNEL,
    HB_CHANNEL,
    IOPUB_CHANNEL,
    SHELL_CHANNEL,
    STDIN_CHANNEL,
)


class KernelListener:
    def __init__(self, client, kernel_id):
        self.client = client
        self.kernel_id = kernel_id
        self.should_listen = True
        self.iopub_signal = signal(IOPUB_CHANNEL)
        self.shell_signal = signal(SHELL_CHANNEL)
        self.stdin_signal = signal(STDIN_CHANNEL)
        self.hb_signal = signal(HB_CHANNEL)
        self.control_signal = signal(CONTROL_CHANNEL)

        # stop listening if the fastapi app shuts down
        signal(APP_SHUTDOWN).connect(self.shutdown_listener)

    async def shutdown_listener(self, *args, **kwargs):
        self.should_listen = False

    async def listen(self):
        while self.should_listen:
            if await self.client.iopub_channel.msg_ready():
                iopub_msg = await self.client.get_iopub_msg()
                self.iopub_signal.send(
                    msg=iopub_msg, kernel_id=self.kernel_id, channel="iopub"
                )

            if await self.client.shell_channel.msg_ready():
                shell_msg = await self.client.get_shell_msg()
                self.shell_signal.send(
                    msg=shell_msg, kernel_id=self.kernel_id, channel="shell"
                )

            if await self.client.stdin_channel.msg_ready():
                stdin_msg = await self.client.get_stdin_msg()
                self.stdin_signal.send(
                    msg=stdin_msg, kernel_id=self.kernel_id, channel="stdin"
                )

            if await self.client.control_channel.msg_ready():
                control_msg = await self.client.get_control_msg()
                self.control_signal.send(
                    msg=control_msg,
                    kernel_id=self.kernel_id,
                    channel="control",
                )

            if await self.client.control_channel.msg_ready():
                control_msg = await self.client.get_hb_msg()
                self.control_signal.send(
                    msg=control_msg, kernel_id=self.kernel_id, channel="hb"
                )

            await asyncio.sleep(0.1)



from typing import Dict

from fastapi.testclient import TestClient


class TestKernels:
    def assert_zsh_spec(self, kern):
        assert kern["kernel_name"] == "zsh"
        assert kern["resource_dir"][-11:] == "kernels/zsh"
        spec = kern["spec"]
        assert "python" in spec["argv"][0]
        assert spec["argv"][1:] == [
            "-m",
            "zsh_jupyter_kernel",
            "-f",
            "{connection_file}",
        ]
        assert spec["env"] == {}
        assert spec["metadata"] == {}
        assert spec["display_name"] == "Z shell"
        assert spec["language"] == "zsh"
        assert spec["interrupt_mode"] == "signal"

    def assert_python3_spec(self, kern):
        assert kern["kernel_name"] == "python3"
        assert kern["resource_dir"][-15:] == "kernels/python3"
        spec = kern["spec"]
        assert "python" in spec["argv"][0]
        assert spec["argv"][1:] == [
            "-m",
            "ipykernel_launcher",
            "-f",
            "{connection_file}",
        ]
        assert spec["env"] == {}
        assert spec["metadata"] == {}
        assert spec["display_name"] == "Python 3"
        assert spec["language"] == "python"
        assert spec["interrupt_mode"] == "signal"

    def assert_bash_spec(self, kern):
        assert kern["kernel_name"] == "bash"
        assert kern["resource_dir"][-12:] == "kernels/bash"
        spec = kern["spec"]
        assert "python" in spec["argv"][0]
        assert spec["argv"][1:] == [
            "-m",
            "bash_kernel",
            "-f",
            "{connection_file}",
        ]
        assert spec["env"] == {"PS1": "$"}
        assert spec["metadata"] == {}
        assert spec["display_name"] == "Bash"
        assert spec["language"] == "bash"
        assert spec["interrupt_mode"] == "signal"

    def assert_r_spec(self, kern):
        assert kern["kernel_name"] == "ir"
        assert kern["resource_dir"][-10:] == "kernels/ir"
        spec = kern["spec"]
        assert "/R" in spec["argv"][0]
        assert spec["argv"][1:] == [
            "--slave",
            "-e",
            "IRkernel::main()",
            "--args",
            "{connection_file}",
        ]
        assert spec["env"] == {}
        assert spec["metadata"] == {}
        assert spec["display_name"] == "R"
        assert spec["language"] == "R"
        assert spec["interrupt_mode"] == "signal"

    def test_specs(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        response = client.get(
            "/kernels/specs", headers=superuser_token_headers
        )
        assert response.status_code == 200
        specs = response.json()["kernel_specs"]
        assert len(specs) >= 4

        for kern in specs:
            if kern["kernel_name"] == "zsh":
                self.assert_zsh_spec(kern)
            elif kern["kernel_name"] == "python3":
                self.assert_python3_spec(kern)
            elif kern["kernel_name"] == "bash":
                self.assert_bash_spec(kern)
            elif kern["kernel_name"] == "ir":
                self.assert_r_spec(kern)


class TestKernelsPermissions:
    def test_specs(
        self,
        client: TestClient,
        readonly_token_headers: Dict[str, str],
        permissionless_token_headers: Dict[str, str],
    ):
        response = client.get(
            "/kernels/specs", headers=permissionless_token_headers
        )
        assert response.status_code == 403
        response = client.get("/kernels/specs", headers=readonly_token_headers)
        assert response.status_code == 200



import asyncio
import json
import os
from datetime import datetime
from enum import Enum, auto
from pathlib import Path
from typing import Any, Dict, List, Optional
from uuid import UUID

import nbformat  # type: ignore
from asyncblink import signal  # type: ignore
from fastapi.logger import logger

from jupyter_d1.commands import execute_d1_command
from jupyter_d1.dav import dav_manager
from jupyter_d1.signals import (
    CELL_ADDED,
    CELL_DELETED,
    CELL_EXECUTION_INPUT,
    CELL_EXECUTION_REPLY,
    CELL_EXECUTION_REQUEST,
    CELL_STREAM,
    CELL_UPDATE,
    COMPLETE_REPLY,
    CONTROL_CHANNEL,
    HB_CHANNEL,
    HISTORY_REPLY,
    IOPUB_CHANNEL,
    KERNEL_INTERRUPTED,
    KERNEL_RESTARTED,
    METADATA_UPDATE,
    NOTEBOOKS_UPDATED,
    PAYLOAD_PAGE,
    SCRATCH_UPDATE,
    SHELL_CHANNEL,
    STDIN_CHANNEL,
    VAR_DETAILS,
    VARS_UPDATE,
)

from ..d1_response import D1Encoder, d1_pairs_hook
from ..kernels import VarsManager, WorkDirManager, get_kernel_definition
from ..models.cell import CellType, CellUpdate
from ..models.execution_state import ExecutionState
from ..models.kernel_variable import KernelVariable
from ..undo_manager import UndoManager
from ..utils import NotebookNode
from .kernel_manager import KernelManager
from .notebook_validator import NotebookValidator

# NBFormat docs
# https://nbformat.readthedocs.io/en/latest/api.html#notebooknode-objects


D1_COMMAND_DELIMITER = "___callisto_d1_command___"


class NBMErrorCode(str, Enum):
    notebook_not_found = "NOTEBOOK_NOT_FOUND"
    cell_not_found = "CELL_NOT_FOUND"


class NBMException(Exception):
    def __init__(
        self,
        error_code: NBMErrorCode,
        a_uuid: Optional[UUID],
        message: Optional[str] = None,
    ):
        if message is None:
            if error_code == NBMErrorCode.notebook_not_found:
                message = f"Notebook not found with uuid: {a_uuid}"
            elif error_code == NBMErrorCode.cell_not_found:
                message = f"Cell not found with uuid: {a_uuid}"
        self.message = message
        self.error_code = error_code
        self.error_uuid = a_uuid


class CommandType(Enum):
    CELL_EXECUTION = auto()
    SCRATCH_EXECUTION = auto()
    VARS_REQUEST = auto()
    VAR_DETAIL_REQUEST = auto()
    UPDATE_WORKDIR_REQUEST = auto()
    CHANGE_WORKDIR_REQUEST = auto()
    COMPLETE_REQUEST = auto()
    HISTORY_REQUEST = auto()


class NotebookCommand:
    def __init__(self, command_type: CommandType, notebook_id: UUID, **kwargs):
        self.type = command_type
        self.notebook_id = notebook_id
        self.extras = kwargs


class Notebook:
    def __init__(
        self,
        node: NotebookNode,
        vars_manager: Optional[VarsManager],
        workdir_manager: Optional[WorkDirManager],
        autosave: bool = True,
    ):
        self.node: NotebookNode = node
        self.umanager = UndoManager()
        self.vars_manager = vars_manager
        self.workdir_manager = workdir_manager
        self.autosave = autosave


class NotebookManager:
    def __init__(self, kmanager: KernelManager):
        self.notebooks: Dict[UUID, Notebook] = {}
        self.kmanager: KernelManager = kmanager
        self.validator: NotebookValidator = NotebookValidator()
        self.last_idle = datetime.utcnow()

        # mapping of message id to command
        self.msg_to_command: Dict[str, NotebookCommand] = {}

        signal(IOPUB_CHANNEL).connect(self.receive_channel_message)
        signal(SHELL_CHANNEL).connect(self.receive_channel_message)
        signal(STDIN_CHANNEL).connect(self.receive_channel_message)
        signal(HB_CHANNEL).connect(self.receive_channel_message)
        signal(CONTROL_CHANNEL).connect(self.receive_channel_message)

    def undo(self, nb_uuid: UUID):
        self.notebooks[nb_uuid].umanager.undo()

    def redo(self, nb_uuid: UUID):
        self.notebooks[nb_uuid].umanager.redo()

    async def create_notebook(
        self,
        kspec_name: Optional[str] = None,
        directory: Optional[str] = None,
        filename: Optional[str] = None,
        working_directory: Optional[str] = None,
    ) -> NotebookNode:
        notebook = nbformat.v4.new_notebook()

        await self.add_notebook(
            notebook,
            kspec_name=kspec_name,
            directory=directory,
            filename=filename,
            working_directory=working_directory,
        )
        return notebook

    async def open_notebook(
        self,
        filepath: str,
        kspec_name: Optional[str] = None,
        working_directory: Optional[str] = None,
    ) -> UUID:
        with open(filepath, "rb") as f:
            nb_content = f.read()
        path = Path(filepath)
        directory = str(path.parent)
        filename = str(path.parts[-1])
        if filename.endswith(".ipynb"):
            filename = filename.replace(".ipynb", "")
        return await self.add_notebook_json(
            nb_content,
            directory=directory,
            kspec_name=kspec_name,
            filename=filename,
            working_directory=working_directory,
        )

    def save_notebook(
        self,
        notebook: Optional[NotebookNode] = None,
        uuid: Optional[UUID] = None,
    ):
        nb = notebook
        if nb is None and uuid is not None and uuid in self.notebooks:
            nb = self.notebooks[uuid].node
        if nb is None:
            raise NBMException(NBMErrorCode.notebook_not_found, uuid)
        path = nb.metadata.jupyter_d1.path
        nbformat.write(nb, path, default=D1Encoder().default)

    def save_notebook_as(
        self,
        uuid: UUID,
        directory: str,
        filename: str,
        delete_old: bool = False,
    ):
        """
        Change save location of the notebook. If delete_old is set,
        the old path of the notebook is deleted (used for renaming).

        This method does not check if the path already exists and
        will overwrite it.
        """
        nb = self.notebooks[uuid]
        if nb is None:
            raise NBMException(NBMErrorCode.notebook_not_found, uuid)
        old_path = nb.node.metadata.jupyter_d1.path
        self.validator.validate(
            nb.node, directory, filename,
        )
        self.save_notebook(nb.node)

        if delete_old:
            os.remove(old_path)

        signal(METADATA_UPDATE).send(
            notebook_id=uuid, metadata=nb.node.metadata
        )

        return nb.node.metadata.jupyter_d1.path

    def upload_notebook(
        self, notebook_data: bytes, directory: str, filename: str
    ):
        nb = nbformat.reads(
            notebook_data, as_version=4, object_pairs_hook=d1_pairs_hook
        )
        self.validator.validate(
            nb, directory, filename, new_notebook_uuid=True
        )
        self.save_notebook(nb)

    async def delete_notebooks(self):
        await self.kmanager.shutdown_all()
        self.notebooks = {}
        dav_manager.clear_providers()
        signal(NOTEBOOKS_UPDATED).send(notebooks=self.get_notebook_nodes())

    async def delete_notebook(self, uuid: UUID):
        self.get_notebook(uuid)  # verify the notebook exists
        await self.kmanager.shutdown_kernel(uuid)
        self.notebooks.pop(uuid, None)
        dav_manager.remove_provider(uuid)
        signal(NOTEBOOKS_UPDATED).send(notebooks=self.get_notebook_nodes())

    def get_notebook_nodes(self) -> Dict[UUID, NotebookNode]:
        return {uuid: nb.node for (uuid, nb) in self.notebooks.items()}

    def get_vars_manager(self, nb_uuid: UUID) -> Optional[VarsManager]:
        nb = self.notebooks.get(nb_uuid)
        return nb.vars_manager if nb is not None else None

    def get_workdir_manager(self, nb_uuid: UUID) -> Optional[WorkDirManager]:
        nb = self.notebooks.get(nb_uuid)
        return nb.workdir_manager if nb is not None else None

    def get_notebooks_json(self) -> str:
        nb_dicts = list(map(lambda x: x.node.dict(), self.notebooks.values()))
        return json.dumps(nb_dicts, cls=D1Encoder)

    def get_notebooks_dicts(self) -> List[Dict[str, Any]]:
        return list(map(lambda x: x.node.dict(), self.notebooks.values()))

    def get_notebook(self, uuid: UUID) -> Notebook:
        nb = self.notebooks.get(uuid)
        if nb is None:
            raise NBMException(NBMErrorCode.notebook_not_found, uuid)
        return nb

    def get_notebook_node(self, uuid: UUID) -> NotebookNode:
        return self.get_notebook(uuid).node

    def get_notebook_json(self, uuid: UUID) -> str:
        node = self.get_notebook(uuid).node
        return nbformat.writes(node, default=D1Encoder().default)

    def get_notebook_dict(self, uuid: UUID) -> Dict[str, Any]:
        return self.get_notebook(uuid).node.dict()

    async def add_notebook(
        self,
        notebook_node: NotebookNode,
        kspec_name: Optional[str] = None,
        directory: Optional[str] = None,
        filename: Optional[str] = None,
        working_directory: Optional[str] = None,
        autosave: bool = True,
    ) -> UUID:
        notebook = None
        try:
            uuid = notebook_node.metadata.jupyter_d1.uuid
            notebook = self.notebooks[uuid]
        except Exception:
            pass
        if notebook is not None:
            return uuid

        resolved_workdir = (
            str(Path(working_directory).resolve())
            if working_directory is not None
            else None
        )

        self.validator.validate(
            notebook_node,
            directory,
            filename,
            working_directory=resolved_workdir,
            new_notebook_uuid=True,
            reset_state=True,
        )
        nb_uuid = notebook_node.metadata.jupyter_d1.uuid

        # Find the spec name
        if kspec_name is None:
            try:
                kspec_name = notebook_node.metadata.kernelspec.name
            except AttributeError:
                kspec_name = None
        available_names = self.kmanager.kernel_names()
        if kspec_name in available_names:
            working_kspec_name = kspec_name
        else:
            # if there's no matching kernel_name, default to a `python` kernel
            working_kspec_name = list(
                filter(lambda x: "python" in x, available_names)
            ).pop()
        working_kspec_name, kernelspec = self.kmanager.get_kernelspec_by_name(
            kspec_name
        )

        work_dir = notebook_node.metadata.jupyter_d1.working_directory

        kernel_definition = get_kernel_definition(
            kernelspec["spec"]["language"]
        )
        vars_manager: Optional[VarsManager] = None
        workdir_manager: Optional[WorkDirManager] = None
        kernel_options = []
        if kernel_definition is not None:
            vars_manager = kernel_definition.create_vars_manager()
            workdir_manager = kernel_definition.create_workdir_manager(
                work_dir
            )
            kernel_options = kernel_definition.kernel_options

        await self.kmanager.start_kernel(
            kernel_name=working_kspec_name,
            uuid=nb_uuid,
            directory=work_dir,
            kernel_options=kernel_options,
        )

        if work_dir is not None:
            dav_manager.add_provider(work_dir, nb_uuid)
        self.notebooks[nb_uuid] = Notebook(
            node=notebook_node,
            vars_manager=vars_manager,
            workdir_manager=workdir_manager,
            autosave=autosave,
        )
        if autosave:
            self.save_notebook(notebook_node)

        signal(CELL_UPDATE).send(
            cells=notebook_node.cells, notebook_id=nb_uuid
        )
        signal(NOTEBOOKS_UPDATED).send(notebooks=self.get_notebook_nodes())
        return nb_uuid

    async def add_notebook_json(
        self,
        notebook_data: bytes,
        kspec_name: Optional[str] = None,
        directory: Optional[str] = None,
        filename: Optional[str] = None,
        working_directory: Optional[str] = None,
        autosave: bool = True,
    ) -> UUID:
        node = nbformat.reads(
            notebook_data, as_version=4, object_pairs_hook=d1_pairs_hook
        )
        return await self.add_notebook(
            node,
            kspec_name=kspec_name,
            directory=directory,
            filename=filename,
            working_directory=working_directory,
            autosave=autosave,
        )

    async def execute(self, notebook_id: UUID, cell_id: UUID):
        notebook = self.get_notebook(notebook_id)
        cell = self.find_cell(notebook_id, cell_id)
        source = cell.source
        uuid = notebook.node.metadata.jupyter_d1.uuid
        msg_id = await self.kmanager.execute(uuid, source)  # returns msg_id
        self.msg_to_command[msg_id] = NotebookCommand(
            CommandType.CELL_EXECUTION,
            uuid,
            cell_id=cell.metadata.jupyter_d1.uuid,
        )
        self.post_execute_request_msg(notebook_id, cell_id, msg_id)
        return msg_id

    def post_execute_request_msg(
        self, notebook_id: UUID, cell_id: UUID, msg_id: str
    ):
        # Repeat the execution request out to any client listening on the
        # websocket so it can keep track of requests from other clients
        signal(CELL_EXECUTION_REQUEST).send(
            notebook_id=notebook_id, cell_id=cell_id, msg_id=msg_id,
        )

    async def execute_scratch(self, notebook_id: UUID, code: str):
        notebook = self.get_notebook(notebook_id)
        uuid = notebook.node.metadata.jupyter_d1.uuid
        msg_id = await self.kmanager.execute(uuid, code, store_history=False)
        self.msg_to_command[msg_id] = NotebookCommand(
            CommandType.SCRATCH_EXECUTION,
            uuid,
            code=code,
            execution_state=ExecutionState.unknown,
        )
        return msg_id

    async def complete(
        self, notebook_id: UUID, code: str, cursor_pos: Optional[int] = None
    ):
        notebook = self.get_notebook(notebook_id)
        uuid = notebook.node.metadata.jupyter_d1.uuid
        msg_id = await self.kmanager.complete(uuid, code, cursor_pos)
        self.msg_to_command[msg_id] = NotebookCommand(
            CommandType.COMPLETE_REQUEST, uuid
        )
        return msg_id

    async def history(self, notebook_id: UUID):
        notebook = self.get_notebook(notebook_id)
        uuid = notebook.node.metadata.jupyter_d1.uuid
        msg_id = await self.kmanager.get_history(uuid)
        self.msg_to_command[msg_id] = NotebookCommand(
            CommandType.HISTORY_REQUEST, uuid
        )
        return msg_id

    async def interrupt_kernel(self, notebook_id: UUID):
        await self.kmanager.interrupt_kernel(notebook_id)
        signal(KERNEL_INTERRUPTED).send(notebook_id=notebook_id,)

    async def perform_restart(self, notebook_id: UUID):
        await self.kmanager.restart_kernel(notebook_id)
        # TODO: Replace this with something smarter, need to wait for kernel
        # to fully start so it's ready for messages
        await asyncio.sleep(2)

    async def restart_kernel(
        self, notebook_id: UUID, clear_output: bool = False
    ):
        await self.perform_restart(notebook_id)
        notebook = self.get_notebook(notebook_id)
        for cell in notebook.node.cells:
            if clear_output:
                if "outputs" in cell:
                    cell.outputs = []

                if "execution_count" in cell:
                    cell.execution_count = None

            cell.metadata.jupyter_d1["execution_state"] = ExecutionState.idle
        signal(KERNEL_RESTARTED).send(
            cells=notebook.node.cells,
            notebook_id=notebook_id,
            run_all_cells=False,
        )

    async def restart_kernel_and_run_all_cells(self, notebook_id: UUID):
        await self.perform_restart(notebook_id)
        notebook = self.get_notebook(notebook_id)
        for cell in notebook.node.cells:
            if "outputs" in cell:
                cell.outputs = []

            if "execution_count" in cell:
                cell.execution_count = None

            if CellType(cell.cell_type) == CellType.code:
                cell.metadata.jupyter_d1[
                    "execution_state"
                ] = ExecutionState.busy

        signal(KERNEL_RESTARTED).send(
            cells=notebook.node.cells,
            notebook_id=notebook_id,
            run_all_cells=True,
        )
        msg_ids = []
        for cell in notebook.node.cells:
            if CellType(cell.cell_type) == CellType.code:
                source = cell.source
                uuid = notebook.node.metadata.jupyter_d1.uuid
                msg_id = await self.kmanager.execute(uuid, source)
                msg_ids.append(msg_id)
                self.msg_to_command[msg_id] = NotebookCommand(
                    CommandType.CELL_EXECUTION,
                    uuid,
                    cell_id=cell.metadata.jupyter_d1.uuid,
                )
                self.post_execute_request_msg(
                    notebook_id, cell.metadata.jupyter_d1.uuid, msg_id
                )
        return msg_ids

    async def update_kernel_vars(self, notebook_id: UUID):
        notebook = self.get_notebook(notebook_id)
        uuid = notebook.node.metadata.jupyter_d1.uuid
        vars_manager = notebook.vars_manager
        if vars_manager is None:
            return
        code = vars_manager.get_vars_code()
        msg_id = await self.kmanager.execute(uuid, code, store_history=False)
        self.msg_to_command[msg_id] = NotebookCommand(
            CommandType.VARS_REQUEST, uuid
        )

    async def get_kernel_var_details(
        self, nb_id: UUID, var_name: str
    ) -> Optional[KernelVariable]:
        notebook = self.get_notebook(nb_id)
        uuid = notebook.node.metadata.jupyter_d1.uuid
        vars_manager = notebook.vars_manager
        if vars_manager is None:
            return None
        code = vars_manager.get_single_var_code(var_name)
        msg_id = await self.kmanager.execute(uuid, code, store_history=False)
        self.msg_to_command[msg_id] = NotebookCommand(
            CommandType.VAR_DETAIL_REQUEST, uuid, var_name=var_name
        )
        single_var_details: Optional[KernelVariable] = None
        waiting = True

        def receive_var_details(
            sender: Any,
            notebook_id: UUID,
            var_name_inc: str,
            single_var: KernelVariable,
            **kwargs,
        ):
            if nb_id == notebook_id and var_name_inc == var_name:
                nonlocal single_var_details
                nonlocal waiting
                single_var_details = single_var
                waiting = False

        signal(VAR_DETAILS).connect(receive_var_details)

        count = 0
        sleep_dur = 0.1
        # Wait ~30 seconds
        while waiting and count < (30 / sleep_dur):
            await asyncio.sleep(sleep_dur)
            count += 1

        return single_var_details

    async def update_kernel_workdir(self, notebook_id: UUID):
        notebook = self.get_notebook(notebook_id)
        uuid = notebook.node.metadata.jupyter_d1.uuid
        workdir_manager = notebook.workdir_manager
        if workdir_manager is None:
            return
        code = workdir_manager.get_cwd_code()
        msg_id = await self.kmanager.execute(uuid, code, store_history=False)
        self.msg_to_command[msg_id] = NotebookCommand(
            CommandType.UPDATE_WORKDIR_REQUEST, uuid
        )

    async def change_kernel_workdir(self, notebook_id: UUID, directory: str):
        notebook = self.get_notebook(notebook_id)
        uuid = notebook.node.metadata.jupyter_d1.uuid
        workdir_manager = notebook.workdir_manager
        if workdir_manager is None:
            return
        code = workdir_manager.get_chdir_code(directory)
        msg_id = await self.kmanager.execute(uuid, code, store_history=False)
        self.msg_to_command[msg_id] = NotebookCommand(
            CommandType.CHANGE_WORKDIR_REQUEST, uuid
        )
        return msg_id

    def change_dav_workdir(self, notebook_id: UUID, directory: str):
        dav_manager.add_provider(directory, notebook_id)

    async def get_vars(self, notebook_id: UUID) -> List[KernelVariable]:
        notebook = self.get_notebook(notebook_id)
        return (
            notebook.vars_manager.vars
            if notebook.vars_manager is not None
            else []
        )

    def find_cell(self, notebook_id: UUID, cell_id: UUID) -> NotebookNode:
        notebook = self.notebooks.get(notebook_id)
        if notebook is None:
            raise NBMException(NBMErrorCode.notebook_not_found, notebook_id)
        for cell in notebook.node.cells:
            if cell.metadata.jupyter_d1.uuid == cell_id:
                return cell
        raise NBMException(NBMErrorCode.cell_not_found, cell_id)

    def delete_cell(self, notebook_id: UUID, cell_id: UUID):
        notebook = self.get_notebook(notebook_id)
        cell = self.find_cell(notebook_id, cell_id)

        # Save cell info for redo operation
        cell_type = cell.cell_type
        source = cell.source
        try:
            # When undoing, we need to add this cell back
            # _before_ the following cell, so get the next
            # cell's uuid as the saved 'before' param
            before_idx = cell.metadata.jupyter_d1.position + 1
            before = notebook.node.cells[before_idx].metadata.jupyter_d1.uuid
        except IndexError:
            # If the indes is off the end of string, then the
            # cell is the last one, which means 'before' is None
            before = None

        notebook.node.cells.remove(cell)
        self.validator.validate(notebook.node)
        signal(CELL_DELETED).send(
            cells=notebook.node.cells, notebook_id=notebook_id
        )

        kwargs = {
            "notebook_id": notebook_id,
            "cell_id": cell_id,
            "cell_type": cell_type,
            "source": source,
            "before": before,
        }
        notebook.umanager.add_action(
            self.create_cell, kwargs=kwargs, name="Delete Cell"
        )

    def update_cell(
        self, notebook_id: UUID, cell_id: UUID, cell_update: CellUpdate
    ) -> NotebookNode:
        notebook = self.get_notebook(notebook_id)
        cell = self.find_cell(notebook_id, cell_id)
        # save info for undo
        undo_cell_update = CellUpdate(
            source=cell.source, cell_type=cell.cell_type
        )

        # update with new info
        if cell_update.source is not None:
            cell.source = cell_update.source
        if cell_update.cell_type is not None:
            cell.cell_type = cell_update.cell_type

        signal(CELL_UPDATE).send(cells=[cell], notebook_id=notebook_id)

        kwargs = {
            "notebook_id": notebook_id,
            "cell_id": cell_id,
            "cell_update": undo_cell_update,
        }
        notebook.umanager.add_action(
            self.update_cell, kwargs=kwargs, name="Update Cell"
        )
        return cell

    def move_cell(
        self, notebook_id: UUID, cell_id: UUID, before: Optional[UUID] = None
    ):
        notebook = self.get_notebook(notebook_id)
        moving_cell = self.find_cell(notebook_id, cell_id)

        # Find the cell before the moving cell and save
        # it for undoing this move op
        undo_before_uuid = None
        cell_index = notebook.node.cells.index(moving_cell) + 1
        if cell_index < len(notebook.node.cells):
            # Otherwise, we're already in the last position
            # so undo_before_uuid should be None
            undo_before_uuid = notebook.node.cells[
                cell_index
            ].metadata.jupyter_d1.uuid

        # Find the before_cell before doing the removal
        # to make sure the cell exists before doing
        # anything destructive
        if before is None:
            before_cell = None
        else:
            before_cell = self.find_cell(notebook_id, before)

        notebook.node.cells.remove(moving_cell)
        if before_cell is None:
            notebook.node.cells.append(moving_cell)
        else:
            position = notebook.node.cells.index(before_cell)
            notebook.node.cells.insert(position, moving_cell)

        # Make sure the new cell has a uuid and positions are renumbered
        self.validator.validate(notebook.node)

        signal(CELL_UPDATE).send(
            cells=notebook.node.cells, notebook_id=notebook_id
        )

        kwargs = {
            "notebook_id": notebook_id,
            "cell_id": cell_id,
            "before": undo_before_uuid,
        }
        notebook.umanager.add_action(
            self.move_cell, kwargs=kwargs, name="Move Cell"
        )

    def create_cell(
        self,
        notebook_id: UUID,
        cell_type: CellType = CellType.code,
        source: str = "",
        before: Optional[UUID] = None,
        cell_id: Optional[UUID] = None,
    ) -> NotebookNode:
        notebook = self.get_notebook(notebook_id)

        if cell_type == CellType.code:
            new_cell = nbformat.v4.new_code_cell(source=source)
        elif cell_type == CellType.markdown:
            new_cell = nbformat.v4.new_markdown_cell(source=source)
        elif cell_type == CellType.raw:
            new_cell = nbformat.v4.new_raw_cell(source=source)
        else:
            new_cell = nbformat.v4.new_code_cell(source=source)

        if before is None:
            notebook.node.cells.append(new_cell)
        else:
            before_cell = self.find_cell(notebook_id, before)
            position = notebook.node.cells.index(before_cell)
            notebook.node.cells.insert(position, new_cell)

        if cell_id is not None:
            if new_cell.metadata.get("jupyter_d1") is None:
                new_cell.metadata.jupyter_d1 = {}
            new_cell.metadata.jupyter_d1.uuid = cell_id

        # Make sure the new cell has a uuid and positions are renumbered
        self.validator.validate(notebook.node)

        new_cell_id = new_cell.metadata.jupyter_d1.uuid

        signal(CELL_ADDED).send(
            cells=notebook.node.cells, notebook_id=notebook_id
        )

        kwargs = {
            "notebook_id": notebook_id,
            "cell_id": new_cell_id,
        }
        notebook.umanager.add_action(
            self.delete_cell, kwargs=kwargs, name="Create Cell"
        )
        return new_cell

    def merge_cells(
        self,
        notebook_id: UUID,
        cell_id: UUID,
        above: bool = False,
        insert_new_line: bool = True,
    ):
        notebook = self.get_notebook(notebook_id)
        merge_cell = self.find_cell(notebook_id, cell_id)

        merge_cell_idx = notebook.node.cells.index(merge_cell)
        source_template = "%s\n%s" if insert_new_line else "%s%s"
        if above:
            cell_index = merge_cell_idx - 1
            if cell_index >= 0:
                merging_cell = notebook.node.cells[cell_index]
            else:
                return
            split_location = len(merging_cell.source)
            merge_cell.source = source_template % (
                merging_cell.source,
                merge_cell.source,
            )
        else:
            cell_index = merge_cell_idx + 1
            if cell_index < len(notebook.node.cells):
                merging_cell = notebook.node.cells[cell_index]
            else:
                return
            split_location = len(merge_cell.source)
            merge_cell.source = source_template % (
                merge_cell.source,
                merging_cell.source,
            )
        split_cell_type = merging_cell.cell_type

        notebook.node.cells.remove(merging_cell)

        # Make sure the new cell has a uuid and positions are renumbered
        self.validator.validate(notebook.node)

        signal(CELL_DELETED).send(
            cells=notebook.node.cells, notebook_id=notebook_id
        )

        kwargs = {
            "notebook_id": notebook_id,
            "cell_id": cell_id,
            "split_location": split_location,
            "split_cell_type": CellType(split_cell_type),
            "above": above,
            "remove_new_line": insert_new_line,
        }
        notebook.umanager.add_action(
            self.split_cell, kwargs=kwargs, name="Merge Cells"
        )

    def split_cell(
        self,
        notebook_id: UUID,
        cell_id: UUID,
        split_location: int,
        split_cell_type: Optional[CellType] = None,
        above: bool = False,
        remove_new_line: bool = False,
    ):
        notebook = self.get_notebook(notebook_id)
        split_cell = self.find_cell(notebook_id, cell_id)

        split_cell_idx = notebook.node.cells.index(split_cell)
        top_source = split_cell.source[:split_location]
        bot_source = split_cell.source[split_location:]

        if remove_new_line and bot_source.startswith("\n"):
            bot_source = bot_source[1:]

        if above:
            new_cell_index = split_cell_idx
            new_cell_source = top_source
            split_cell_source = bot_source
        else:
            new_cell_index = split_cell_idx + 1
            new_cell_source = bot_source
            split_cell_source = top_source

        if split_cell_type is None:
            split_cell_type = CellType(split_cell.cell_type)
        if split_cell_type == CellType.code:
            new_cell = nbformat.v4.new_code_cell(source=new_cell_source)
        elif split_cell_type == CellType.markdown:
            new_cell = nbformat.v4.new_markdown_cell(source=new_cell_source)
        elif split_cell_type == CellType.raw:
            new_cell = nbformat.v4.new_raw_cell(source=new_cell_source)
        else:
            new_cell = nbformat.v4.new_code_cell(source=new_cell_source)

        split_cell.source = split_cell_source

        if new_cell_index >= len(notebook.node.cells):
            notebook.node.cells.append(new_cell)
        else:
            notebook.node.cells.insert(new_cell_index, new_cell)

        # Make sure the new cell has a uuid and positions are renumbered
        self.validator.validate(notebook.node)

        signal(CELL_UPDATE).send(
            cells=notebook.node.cells, notebook_id=notebook_id
        )

        kwargs = {
            "notebook_id": notebook_id,
            "cell_id": cell_id,
            "above": above,
            "insert_new_line": remove_new_line,
        }
        notebook.umanager.add_action(
            self.merge_cells, kwargs=kwargs, name="Split Cells"
        )

    def extract_d1_commands(self, output: NotebookNode) -> NotebookNode:
        splits = output.text.split(D1_COMMAND_DELIMITER)
        split_len = len(splits)
        if split_len > 2:
            if split_len % 2 == 0:
                new_text = "".join(splits[::2] + [splits[-1]])
                commands = splits[1:-1:2]
            else:
                new_text = "".join(splits[::2])
                commands = splits[1::2]
            output.text = new_text
            for command in commands:
                execute_d1_command(command)
        return output

    async def receive_channel_message(
        self, sender, msg, kernel_id, channel, **kwargs
    ):
        logger.debug("nb_manager received channel message")
        parent_header = msg.get("parent_header")
        msg_type = msg.get("msg_type")
        content = msg.get("content")
        if parent_header is None:
            logger.debug("Can't get parent header")
            return

        parent_id = parent_header.get("msg_id")
        if parent_id is None:
            logger.debug("Can't get parent id")
            return

        command = self.msg_to_command.get(parent_id)
        if command is None:
            logger.debug(f"Msg ({parent_id}) not found in msg_to_command:")
            logger.debug(f"    {self.msg_to_command}")
            return

        nb_id = command.notebook_id
        if command.type == CommandType.CELL_EXECUTION:
            cell_id = command.extras["cell_id"]
            cell = self.find_cell(command.notebook_id, cell_id)

        if msg_type == "execute_reply" and content is not None:
            if command.type == CommandType.CELL_EXECUTION:
                execution_count = content["execution_count"]
                cell.execution_count = execution_count
                logger.debug(
                    f"nb_manager posting CELL_EXECUTION_REPLY - execute_reply"
                )
                signal(CELL_EXECUTION_REPLY).send(
                    cell_id=cell_id,
                    execution_count=execution_count,
                    notebook_id=nb_id,
                    parent_id=parent_id,
                )
            if (
                "status" in content
                and content["status"] == "ok"
                and "payload" in content
            ):
                for payload_item in content["payload"]:
                    if payload_item["source"] == "page":
                        signal(PAYLOAD_PAGE).send(
                            msg_id=parent_id,
                            notebook_id=nb_id,
                            data=payload_item["data"],
                            start=payload_item["start"],
                        )
        elif msg_type == "execute_input":
            if command.type == CommandType.CELL_EXECUTION:
                signal(CELL_EXECUTION_INPUT).send(
                    notebook_id=nb_id, parent_id=parent_id, content=content,
                )
        elif (
            msg_type == "complete_reply"
            and command.type == CommandType.COMPLETE_REQUEST
            and content is not None
            and content.get("status") == "ok"
        ):
            logger.debug(f"nb_manager posting COMPLETE_REPLY - {msg_type}")
            signal(COMPLETE_REPLY).send(
                msg_id=parent_id,
                notebook_id=nb_id,
                matches=content["matches"],
                cursor_start=content["cursor_start"],
                cursor_end=content["cursor_end"],
            )
        elif msg_type == "history_reply" and content is not None:
            logger.debug(f"nb_manager posting HISTORY_REPLY - {msg_type}")
            signal(HISTORY_REPLY).send(
                msg_id=parent_id, notebook_id=nb_id, history=content["history"]
            )
        elif msg_type in ("execute_result", "error", "stream", "display_data"):
            output = nbformat.v4.output_from_msg(msg)
            if msg_type == "stream":
                # Suppress the stream output for internal commands
                suppress_stream_commands = [
                    CommandType.UPDATE_WORKDIR_REQUEST,
                    CommandType.VARS_REQUEST,
                    CommandType.VAR_DETAIL_REQUEST,
                    CommandType.UPDATE_WORKDIR_REQUEST,
                ]
                if command.type not in suppress_stream_commands:
                    output = self.extract_d1_commands(output)
            if command.type == CommandType.CELL_EXECUTION:
                if "outputs" not in cell:
                    cell.outputs = []
                cell.outputs.append(output)
                if "execution_count" in output:
                    cell.execution_count = output.execution_count
                logger.debug(f"nb_manager posting CELL_UPDATE - {msg_type}")
                if msg_type == "stream":
                    signal(CELL_STREAM).send(
                        output=output, notebook_id=nb_id, cell_id=cell_id
                    )
                else:
                    signal(CELL_UPDATE).send(cells=[cell], notebook_id=nb_id)
            elif command.type == CommandType.SCRATCH_EXECUTION:
                logger.debug(f"nb_manager posting SCRATCH_UPDATE - {msg_type}")
                signal(SCRATCH_UPDATE).send(
                    msg_id=parent_id, notebook_id=nb_id, output=output
                )
            elif command.type == CommandType.VARS_REQUEST:
                vars_manager = self.get_vars_manager(nb_id)
                if (
                    output.output_type in ("stream", "display_data")
                    and vars_manager is not None
                ):
                    vars_manager.parse_output(output)
            elif command.type == CommandType.VAR_DETAIL_REQUEST:
                vars_manager = self.get_vars_manager(nb_id)
                if (
                    output.output_type in ("stream", "display_data")
                    and vars_manager is not None
                ):
                    single_var = vars_manager.parse_single_var_response(output)
                    signal(VAR_DETAILS).send(
                        notebook_id=nb_id,
                        var_name_inc=command.extras.get("var_name", None),
                        single_var=single_var,
                    )
            elif command.type in (
                CommandType.UPDATE_WORKDIR_REQUEST,
                CommandType.CHANGE_WORKDIR_REQUEST,
            ):
                workdir_manager = self.get_workdir_manager(nb_id)
                if (
                    output.output_type in ("stream", "display_data")
                    and workdir_manager is not None
                ):
                    if command.type == CommandType.UPDATE_WORKDIR_REQUEST:
                        workdir_manager.parse_cwd_output(output)
                    elif command.type == CommandType.CHANGE_WORKDIR_REQUEST:
                        workdir_manager.parse_chdir_output(output)

        if msg_type == "status" and content is not None:

            execution_state = content.get(
                "execution_state"
            )  # busy/idle/starting
            if command.type == CommandType.CELL_EXECUTION:
                cell.metadata.jupyter_d1["execution_state"] = execution_state
                logger.debug("nb_manager posting CELL_UPDATE - status")
                if execution_state == ExecutionState.busy:
                    cell.outputs = []
                signal(CELL_UPDATE).send(cells=[cell], notebook_id=nb_id)
                if execution_state == ExecutionState.idle:
                    notebook = self.notebooks.get(nb_id, None)
                    if notebook is not None and notebook.autosave:
                        self.save_notebook(uuid=nb_id)
            elif command.type == CommandType.SCRATCH_EXECUTION:
                logger.debug("nb_manager posting SCRATCH_UPDATE - status")
                command.extras["execution_state"] = ExecutionState(
                    execution_state
                )
                signal(SCRATCH_UPDATE).send(
                    msg_id=parent_id,
                    notebook_id=nb_id,
                    execution_state=execution_state,
                )
            elif command.type == CommandType.VARS_REQUEST:
                vars_manager = self.get_vars_manager(nb_id)
                if vars_manager is not None:
                    if execution_state == ExecutionState.busy:
                        vars_manager.on_request_start()
                    elif execution_state == ExecutionState.idle:
                        logger.debug(
                            f"nb_manager posting VARS_UPDATE - {msg_type}"
                        )
                        self.msg_to_command.pop(parent_id)
                        vars = vars_manager.on_request_end()
                        signal(VARS_UPDATE).send(notebook_id=nb_id, vars=vars)
                return
            elif command.type == CommandType.VAR_DETAIL_REQUEST:
                if execution_state == ExecutionState.idle:
                    self.msg_to_command.pop(parent_id)
                return
            elif command.type in (
                CommandType.UPDATE_WORKDIR_REQUEST,
                CommandType.CHANGE_WORKDIR_REQUEST,
            ):
                workdir_manager = self.get_workdir_manager(nb_id)
                if workdir_manager is not None:
                    if execution_state == ExecutionState.idle:
                        self.msg_to_command.pop(parent_id)
                        metadata = self.get_notebook_node(nb_id).metadata
                        resolved_workdir = str(
                            Path(workdir_manager.workdir).resolve()
                        )
                        if (
                            metadata.jupyter_d1.working_directory
                            != resolved_workdir
                        ):
                            logger.debug(
                                f"nb_manager posting METADATA_UPDATE - "
                                f"{msg_type}"
                            )
                            metadata.jupyter_d1.working_directory = (
                                resolved_workdir
                            )
                            notebook = self.notebooks.get(nb_id, None)
                            if notebook is not None and notebook.autosave:
                                self.save_notebook(uuid=nb_id)
                            self.change_dav_workdir(nb_id, resolved_workdir)
                            signal(METADATA_UPDATE).send(
                                notebook_id=nb_id, metadata=metadata
                            )

            # If something just finished running, update kernel vars
            if execution_state == ExecutionState.idle:
                self.last_idle = datetime.utcnow()

                if command.type not in (
                    CommandType.UPDATE_WORKDIR_REQUEST,
                    CommandType.CHANGE_WORKDIR_REQUEST,
                ):
                    await self.update_kernel_workdir(nb_id)
                    await self.update_kernel_vars(nb_id)

    def is_idle(self):
        for notebook in self.notebooks.values():
            for cell in notebook.node.cells:
                if (
                    "execution_state" in cell.metadata.jupyter_d1
                    and cell.metadata.jupyter_d1.execution_state != "idle"
                ):
                    return False
        for command in self.msg_to_command.values():
            if command.extras.get("execution_state", None) not in (
                ExecutionState.idle,
                None,
            ):
                return False

        return True



from wsgidav.dc.base_dc import BaseDomainController  # type: ignore

from .deps import extract_permission


class JWTDomainController(BaseDomainController):
    def get_domain_realm(self, path_info, environ):
        """Resolve a relative url to the appropriate realm name."""
        realm = self._calc_realm_from_path_provider(path_info, environ)
        # This solves a bug in send_basic_auth_response in HTTPAuthenticator
        # where it tries to concatenate realm to a string and fails if
        # it's None
        return realm if realm is not None else "None"

    def require_authentication(self, realm, environ):
        """Return True if this realm requires authentication."""
        return True

    def basic_auth_user(self, realm, user_name, password, environ):
        """Returns True if this user_name/password pair is valid for the 
        realm, False otherwise. Used for basic authentication."""
        try:
            permission = extract_permission(user_name)
            roles = ()
            permissions = ()
            if permission.read_access is True:
                roles = ("reader",)
                permissions = ("browse_dir",)
            if permission.write_access is True:
                roles = ("admin", "editor", "reader")
                permissions = (
                    "browse_dir",
                    "delete_resource",
                    "edit_resource",
                )
            environ["wsgidav.auth.roles"] = roles
            environ["wsgidav.auth.permissions"] = permissions
            return (
                permission.read_access is True
                or permission.write_access is True
            )
        except Exception as e:
            print(e)
            return False

    def supports_http_digest_auth(self):
        return False



import json
from typing import List, Optional

from fastapi.logger import logger

from ...models.kernel_variable import KernelVariable
from ...settings import settings
from ...utils import NotebookNode
from ..vars_manager import VarsManager

# Prints list of current vars to stdout.
# Strange var names used to avoid affecting users
# variables.
raw_get_vars_code = """
try:
    from callisto import format_vars as __d1_format_vars
    print(__d1_format_vars(vars(), abbrev_len={abbrev_len}))
except Exception as e:
    print(e)
"""  # noqa

raw_get_single_var_code = """
try:
    from callisto import format_var as __d1_format_var
    print(__d1_format_var({var_name}, "{var_name}", abbrev_len=None))
except Exception as e:
    print(e)
"""  # noqa


class PythonVarsManager(VarsManager):
    def get_vars_code(self) -> str:
        return raw_get_vars_code.format(abbrev_len=settings.VAR_ABBREV_LEN)

    def get_single_var_code(self, var_name: str) -> str:
        return raw_get_single_var_code.format(var_name=var_name)

    def parse_vars_response(
        self, vars_response: NotebookNode
    ) -> List[KernelVariable]:
        vars: List[KernelVariable] = []
        if "text" not in vars_response:
            return vars
        try:
            json_vars = json.loads(vars_response.text)
            for json_var in json_vars:
                vars.append(
                    KernelVariable(
                        name=json_var.get("name"),
                        type=json_var.get("type"),
                        abbreviated=json_var.get("abbreviated"),
                        value=json_var.get("value"),
                        summary=json_var.get("summary"),
                    )
                )
        except Exception as e:
            logger.debug(
                f"Exception parsing vars for python kernel: {e}, "
                f"{vars_response.text}"
            )
        return vars

    def parse_single_var_response(
        self, var_response: NotebookNode
    ) -> Optional[KernelVariable]:
        var = None
        if "text" not in var_response:
            return var
        try:
            json_var = json.loads(var_response.text)
            var = KernelVariable(
                name=json_var.get("name"),
                type=json_var.get("type"),
                abbreviated=json_var.get("abbreviated"),
                value=json_var.get("value"),
                summary=json_var.get("summary"),
            )
        except Exception as e:
            logger.debug(
                f"Exception parsing var for python kernel: {e}, "
                f"{var_response.text}"
            )
        return var



import asyncio
from typing import List, Type

import nbformat  # type: ignore
from asyncblink import signal  # type: ignore
from fastapi import WebSocket
from fastapi.logger import logger
from websockets.exceptions import ConnectionClosedError

from .signals import APP_SHUTDOWN

NotebookNode = Type[nbformat.notebooknode.NotebookNode]


def add_to_node_metadata(node: NotebookNode, **kwargs):
    if node.metadata.hasattr("jupyter_d1") is False:
        node.metadata.jupyter_d1 = {}
    node.metadata.jupyter_d1.update(kwargs)


def add_to_dict_metadata(someDict, **kwargs):
    if someDict.get("metadata") is None:
        someDict["metadata"] = {}
    metadata = someDict["metadata"]

    if metadata.get("jupyter_d1") is None:
        metadata["jupyter_d1"] = {}
    jupyter_d1 = metadata["jupyter_d1"]

    jupyter_d1.update(kwargs)


class BoolObject:
    def __init__(self, value=False):
        self.value = value

    def __bool__(self):
        return self.value


async def websocket_poll(websocket: WebSocket, msgs: List[str]):
    # use an object to communicate across coroutines
    connected = BoolObject(value=True)

    async def shutdown_listener(*args, **kwargs):
        connected.value = False

    signal(APP_SHUTDOWN).connect(shutdown_listener)

    while connected:
        await asyncio.sleep(0.1)
        if len(msgs) != 0:
            msg = msgs.pop(0)
            try:
                await websocket.send_text(msg)
            except ConnectionClosedError:
                logger.debug("client closed connection to websocket service")
                connected.value = False

    await websocket.close()



from datetime import datetime
from typing import List, Optional

from pydantic import BaseModel

from .base_wrapper import BaseWrapper


class DiskStats(BaseModel):
    free: int
    percent: float
    total: int


class MemoryStats(BaseModel):
    available: int
    total: int


class CPUStats(BaseModel):
    percent: float
    load_percent: float


class GPUStats(BaseModel):
    name: Optional[str]
    gpu_usage_percent: Optional[float]
    memory_usage_percent: Optional[float]
    memory_free: Optional[int]
    memory_used: Optional[int]
    memory_total: Optional[int]


class ServerStats(BaseModel):
    timestamp: datetime
    disk: DiskStats
    memory: MemoryStats
    cpu: CPUStats
    gpu: List[GPUStats]


class ServerStatsWrapper(BaseWrapper):
    server_stats: Optional[ServerStats]


class ServerStatsListWrapper(BaseWrapper):
    server_stats: List[ServerStats]



from fastapi import APIRouter

router = APIRouter()


@router.get("")
async def hello():
    return {"message": "Hello World"}


@router.get("/{name}")
async def hello_user(name: str):
    return {"message": f"Hello {name}"}



from ...utils import NotebookNode
from ..workdir_manager import WorkDirManager

raw_import_line = "import os as ___d1os"
raw_chdir_line = "___d1os.chdir('{directory}')"
raw_cwd_line = "print(___d1os.getcwd())"

raw_chdir_code = f"""
{raw_import_line}
{raw_chdir_line}
{raw_cwd_line}
"""
raw_cwd_code = f"""
{raw_import_line}
{raw_cwd_line}"""


class PythonWorkDirManager(WorkDirManager):
    def get_chdir_code(self, directory: str) -> str:
        return raw_chdir_code.format(directory=directory)

    def get_cwd_code(self) -> str:
        return raw_cwd_code

    def parse_chdir_response(self, response: NotebookNode) -> str:
        return response.text.strip()

    def parse_cwd_response(self, response: NotebookNode) -> str:
        return response.text.strip()



import json
import os
import pathlib
from typing import Dict, Tuple

import pytest  # type: ignore
from fastapi.testclient import TestClient

from jupyter_d1.settings import settings


def upload_notebook(
    client: TestClient,
    token_headers: Dict[str, str],
    filename: str = "simple.ipynb",
) -> str:
    nb_filename = f"jupyter_d1/tests/notebooks/{filename}"
    nb_json = open(nb_filename).read()
    response = client.post(
        "/notebooks/upload",
        params={"filename": filename},
        data=nb_json,
        headers=token_headers,
    )
    assert response.status_code == 201
    path = response.json()["path"]

    response = client.get(
        f"/notebooks/open/?filepath={path}", headers=token_headers
    )
    assert response.status_code == 201
    resp_json = response.json()["notebook"]
    uuid = resp_json["metadata"]["jupyter_d1"]["uuid"]
    return uuid


@pytest.mark.usefixtures("clear_notebooks", "clear_notebook_directory")
class TestNotebook:
    def test_notebook(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        uuid = upload_notebook(client, superuser_token_headers)
        response = client.get(
            f"/notebooks/{uuid}", headers=superuser_token_headers
        )
        assert response.status_code == 200
        nb = response.json()["notebook"]
        assert "cells" in nb.keys()
        assert "metadata" in nb.keys()
        assert "nbformat" in nb.keys()
        assert "nbformat_minor" in nb.keys()
        assert nb["metadata"]["jupyter_d1"]["uuid"] == uuid

    def test_notebooks(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        uuid = upload_notebook(client, superuser_token_headers, "simple.ipynb")
        uuid2 = upload_notebook(
            client, superuser_token_headers, "other_simple.ipynb"
        )
        response = client.get("/notebooks", headers=superuser_token_headers)
        assert response.status_code == 200
        resp_json = response.json()
        notebooks = resp_json["notebooks"]
        assert len(notebooks) == 2

        ret_uuid = notebooks[0]["metadata"]["jupyter_d1"]["uuid"]
        ret_uuid2 = notebooks[1]["metadata"]["jupyter_d1"]["uuid"]
        assert uuid != uuid2
        assert set([uuid, uuid2]) == set([ret_uuid, ret_uuid2])

    def test_delete_notebooks(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        upload_notebook(client, superuser_token_headers, "simple.ipynb")
        upload_notebook(client, superuser_token_headers, "other_simple.ipynb")
        response = client.get("/notebooks", headers=superuser_token_headers)
        assert response.status_code == 200
        resp_json = response.json()["notebooks"]
        assert len(resp_json) == 2

        response = client.delete("/notebooks", headers=superuser_token_headers)
        assert response.status_code == 204

        response = client.get("/notebooks", headers=superuser_token_headers)
        assert response.status_code == 200
        resp_json = response.json()["notebooks"]
        assert len(resp_json) == 0

    def test_delete_notebook(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        uuid = upload_notebook(client, superuser_token_headers, "simple.ipynb")
        uuid2 = upload_notebook(
            client, superuser_token_headers, "other_simple.ipynb"
        )
        response = client.get("/notebooks", headers=superuser_token_headers)
        assert response.status_code == 200
        resp_json = response.json()["notebooks"]
        assert len(resp_json) == 2
        ret_uuid = resp_json[0]["metadata"]["jupyter_d1"]["uuid"]
        ret_uuid2 = resp_json[1]["metadata"]["jupyter_d1"]["uuid"]
        assert set([uuid, uuid2]) == set([ret_uuid, ret_uuid2])

        # delete the first one
        response = client.delete(
            f"/notebooks/{uuid}", headers=superuser_token_headers
        )
        assert response.status_code == 204

        response = client.get("/notebooks", headers=superuser_token_headers)
        assert response.status_code == 200
        resp_json = response.json()["notebooks"]
        assert len(resp_json) == 1
        # remaining notebook should have uuid2
        assert resp_json[0]["metadata"]["jupyter_d1"]["uuid"] == uuid2

    def test_cells(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        uuid = upload_notebook(client, superuser_token_headers, "simple.ipynb")

        response = client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]
        assert len(cells) == 4

        for cell in cells:
            uuid = cell["metadata"]["jupyter_d1"]["uuid"]
            assert len(uuid) == 36
            uuid = cell["metadata"]["jupyter_d1"]["notebook_uuid"] == uuid

        cell0 = cells[0]
        assert cell0["cell_type"] == "markdown"
        assert cell0["source"] == "## Simple Test Notebook"

        cell1 = cells[1]
        assert cell1["cell_type"] == "code"
        assert cell1["source"] == 'print("Larry the Llama")'

    def test_one_cell(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        uuid = upload_notebook(client, superuser_token_headers, "simple.ipynb")

        response = client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]

        cell1 = cells[1]
        cell1_uuid = cell1["metadata"]["jupyter_d1"]["uuid"]

        response = client.get(
            f"/notebooks/{uuid}/cells/{cell1_uuid}",
            headers=superuser_token_headers,
        )
        assert response.status_code == 200

        cell = response.json()["cell"]
        uuid = cell["metadata"]["jupyter_d1"]["uuid"]
        assert uuid == cell1_uuid
        assert cell["cell_type"] == "code"
        assert cell["source"] == 'print("Larry the Llama")'

    def test_patch_cell(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        uuid = upload_notebook(client, superuser_token_headers, "simple.ipynb")

        response = client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]

        cell1 = cells[1]
        cell1_uuid = cell1["metadata"]["jupyter_d1"]["uuid"]

        new_source = 'hello = "world"'
        response = client.patch(
            f"/notebooks/{uuid}/cells/{cell1_uuid}",
            headers=superuser_token_headers,
            json={"source": new_source},
        )
        assert response.status_code == 200
        # check that the returned cell has the new source
        cell = response.json()["cell"]
        assert cell["source"] == new_source

        # Assert that change was written to disk
        with open(f"{settings.ROOT_DIR}/simple.ipynb") as f:
            file_nb = json.loads(f.read())
        assert (
            file_nb["cells"][1]["metadata"]["jupyter_d1"]["uuid"]
            == cell["metadata"]["jupyter_d1"]["uuid"]
        )
        assert file_nb["cells"][1]["source"][0] == new_source

        # check that retrieving the cell has has the new source
        response = client.get(
            f"/notebooks/{uuid}/cells/{cell1_uuid}",
            headers=superuser_token_headers,
        )
        assert response.status_code == 200
        cell = response.json()["cell"]
        assert cell["source"] == new_source

    def test_create_cell(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        uuid = upload_notebook(client, superuser_token_headers, "simple.ipynb")

        # get the existing cells
        response = client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]
        orig_uuids = list(
            map(lambda x: x["metadata"]["jupyter_d1"]["uuid"], cells)
        )

        assert len(orig_uuids) == 4
        before_uuid = orig_uuids[1]  # place the new cell before position 1

        # add a new cell before index 1
        src = "__**Hello Callist**__"
        params = {
            "before": before_uuid,
            "cell_type": "markdown",
            "source": src,
        }
        response = client.post(
            f"/notebooks/{uuid}/cells",
            json=params,
            headers=superuser_token_headers,
        )
        assert response.status_code == 201
        new_cell = response.json()["cell"]
        assert new_cell["cell_type"] == "markdown"
        assert new_cell["source"] == src

        new_cell_uuid = new_cell["metadata"]["jupyter_d1"]["uuid"]
        assert len(new_cell_uuid) == 36

        # insert the new uuid into the original list
        orig_uuids.insert(1, new_cell_uuid)

        # Assert that change was written to disk
        with open(f"{settings.ROOT_DIR}/simple.ipynb") as f:
            file_nb = json.loads(f.read())
        assert (
            file_nb["cells"][1]["metadata"]["jupyter_d1"]["uuid"]
            == new_cell["metadata"]["jupyter_d1"]["uuid"]
        )

        # get the new list of cells
        response = client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        new_cells = response.json()["cells"]
        new_uuids = list(
            map(lambda x: x["metadata"]["jupyter_d1"]["uuid"], new_cells)
        )
        assert len(new_uuids) == 5

        # assert the new list of uuids is the old list with the new uuid
        assert new_uuids == orig_uuids

        # add a new cell at the end (omit position parameter)
        response = client.post(
            f"/notebooks/{uuid}/cells",
            json={},
            headers=superuser_token_headers,
        )
        assert response.status_code == 201
        new_cell = response.json()["cell"]

        new_cell_uuid = new_cell["metadata"]["jupyter_d1"]["uuid"]
        assert len(new_cell_uuid) == 36

        # insert the new uuid into the original list
        orig_uuids.append(new_cell_uuid)

        # get the new list of cells
        response = client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        new_cells = response.json()["cells"]
        new_uuids = list(
            map(lambda x: x["metadata"]["jupyter_d1"]["uuid"], new_cells)
        )

        # assert the new list of uuids is the old list with the new uuid
        assert new_uuids == orig_uuids

    def test_move_cell(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        uuid = upload_notebook(client, superuser_token_headers, "simple.ipynb")

        # get the existing cells
        response = client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]
        orig_uuids = list(
            map(lambda x: x["metadata"]["jupyter_d1"]["uuid"], cells)
        )

        assert len(orig_uuids) == 4
        # move the third cell 'before' the second cell (swap positions)
        move_uuid = orig_uuids[2]  # place the new cell before position 1
        before_uuid = orig_uuids[1]  # place the new cell before position 1

        params = {"before": before_uuid}
        response = client.get(
            f"/notebooks/{uuid}/cells/{move_uuid}/move",
            params=params,
            headers=superuser_token_headers,
        )
        assert response.status_code == 204

        # This is the expected reordering
        reordered_uuids = [
            orig_uuids[0],
            orig_uuids[2],  # swap indecies 1, 2
            orig_uuids[1],
            orig_uuids[3],
        ]

        # Assert that change was written to disk
        with open(f"{settings.ROOT_DIR}/simple.ipynb") as f:
            file_nb = json.loads(f.read())
        file_uuids = list(
            map(
                lambda x: x["metadata"]["jupyter_d1"]["uuid"], file_nb["cells"]
            )
        )
        assert file_uuids == reordered_uuids

        # get the new list of cells
        response = client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        new_cells = response.json()["cells"]
        new_uuids = list(
            map(lambda x: x["metadata"]["jupyter_d1"]["uuid"], new_cells)
        )
        assert len(new_uuids) == 4

        # assert the new list of uuids is the old list with the new uuid
        assert new_uuids == reordered_uuids

        move_uuid = reordered_uuids[0]
        # move the first cell to the end (omit position parameter)
        response = client.get(
            f"/notebooks/{uuid}/cells/{move_uuid}/move",
            headers=superuser_token_headers,
        )
        assert response.status_code == 204

        # final list just has the first cell moved to the end
        final_uuids = [
            reordered_uuids[1],
            reordered_uuids[2],
            reordered_uuids[3],
            reordered_uuids[0],
        ]

        # get the new list of cells
        response = client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        new_cells = response.json()["cells"]
        new_uuids = list(
            map(lambda x: x["metadata"]["jupyter_d1"]["uuid"], new_cells)
        )

        # assert the new list of uuids is the old list with the new uuid
        assert new_uuids == final_uuids

    def _test_merge_cells(
        self,
        client: TestClient,
        superuser_token_headers: Dict[str, str],
        uuid: str,
        position: int,
        above: bool,
        cell_types: Tuple[str, str],
    ):
        if above:
            position_1 = position - 1
            position_2 = position
        else:
            position_1 = position
            position_2 = position + 1

        # get the existing cells
        response = client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]
        assert len(cells) == 4
        cell_uuid = cells[position]["metadata"]["jupyter_d1"]["uuid"]
        cell_1_source = cells[position_1]["source"]
        assert cells[position_1]["cell_type"] == cell_types[0]
        cell_2_source = cells[position_2]["source"]
        assert cells[position_2]["cell_type"] == cell_types[1]

        params = {}
        if above:
            params["above"] = True
        response = client.get(
            f"/notebooks/{uuid}/cells/{cell_uuid}/merge",
            params=params,
            headers=superuser_token_headers,
        )
        assert response.status_code == 204

        response = client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]
        assert len(cells) == 3
        cell = cells[position_1]
        assert cell["source"] == f"{cell_1_source}\n{cell_2_source}"
        assert cell["cell_type"] == cell_types[1 if above else 0]
        assert cell["metadata"]["jupyter_d1"]["uuid"] == cell_uuid

        response = client.get(
            f"/notebooks/{uuid}/undo", headers=superuser_token_headers
        )
        assert response.status_code == 204

        response = client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]
        assert len(cells) == 4
        assert cells[position]["metadata"]["jupyter_d1"]["uuid"] == cell_uuid
        assert cells[position_1]["source"] == cell_1_source
        assert cells[position_1]["cell_type"] == cell_types[0]
        assert cells[position_2]["source"] == cell_2_source
        assert cells[position_2]["cell_type"] == cell_types[1]

        response = client.get(
            f"/notebooks/{uuid}/redo", headers=superuser_token_headers
        )
        assert response.status_code == 204

        response = client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]
        assert len(cells) == 3
        cell = cells[position_1]
        assert cell["source"] == f"{cell_1_source}\n{cell_2_source}"
        assert cell["cell_type"] == cell_types[1 if above else 0]
        assert cell["metadata"]["jupyter_d1"]["uuid"] == cell_uuid

    def test_merge_cells(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        uuid = upload_notebook(client, superuser_token_headers, "simple.ipynb")
        self._test_merge_cells(
            client,
            superuser_token_headers,
            uuid=uuid,
            position=1,
            above=False,
            cell_types=("code", "code"),
        )

    def test_merge_cells_above(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        uuid = upload_notebook(client, superuser_token_headers, "simple.ipynb")
        self._test_merge_cells(
            client,
            superuser_token_headers,
            uuid=uuid,
            position=1,
            above=True,
            cell_types=("markdown", "code"),
        )

    def test_merge_first_2_cells(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        uuid = upload_notebook(client, superuser_token_headers, "simple.ipynb")
        self._test_merge_cells(
            client,
            superuser_token_headers,
            uuid=uuid,
            position=0,
            above=False,
            cell_types=("markdown", "code"),
        )

    def test_merge_last_2_cells(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        uuid = upload_notebook(client, superuser_token_headers, "simple.ipynb")
        self._test_merge_cells(
            client,
            superuser_token_headers,
            uuid=uuid,
            position=3,
            above=True,
            cell_types=("code", "code"),
        )

    def _test_split_cell(
        self,
        client: TestClient,
        superuser_token_headers: Dict[str, str],
        uuid: str,
        position: int,
        split_location: int,
        expected_sources: Tuple[str, str],
        cell_type: str,
    ):
        # get the existing cells
        response = client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]
        assert len(cells) == 4
        cell_uuid = cells[position]["metadata"]["jupyter_d1"]["uuid"]
        original_source = cells[position]["source"]
        assert cells[position]["cell_type"] == cell_type

        response = client.get(
            f"/notebooks/{uuid}/cells/{cell_uuid}/split",
            params={"split_location": split_location},
            headers=superuser_token_headers,
        )
        assert response.status_code == 204

        response = client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]
        assert len(cells) == 5
        assert cells[position]["source"] == expected_sources[0]
        assert cells[position]["cell_type"] == cell_type
        assert cells[position + 1]["source"] == expected_sources[1]
        assert cells[position + 1]["cell_type"] == cell_type
        assert cells[position]["metadata"]["jupyter_d1"]["uuid"] == cell_uuid

        response = client.get(
            f"/notebooks/{uuid}/undo", headers=superuser_token_headers
        )
        assert response.status_code == 204

        response = client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]
        assert len(cells) == 4
        assert cells[position]["metadata"]["jupyter_d1"]["uuid"] == cell_uuid
        assert cells[position]["source"] == original_source
        assert cells[position]["cell_type"] == cell_type

        response = client.get(
            f"/notebooks/{uuid}/redo", headers=superuser_token_headers
        )
        assert response.status_code == 204

        response = client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cells = response.json()["cells"]
        assert len(cells) == 5
        assert cells[position]["source"] == expected_sources[0]
        assert cells[position]["cell_type"] == cell_type
        assert cells[position + 1]["source"] == expected_sources[1]
        assert cells[position + 1]["cell_type"] == cell_type
        assert cells[position]["metadata"]["jupyter_d1"]["uuid"] == cell_uuid

    def test_split_cell(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        uuid = upload_notebook(client, superuser_token_headers, "simple.ipynb")
        self._test_split_cell(
            client,
            superuser_token_headers,
            uuid=uuid,
            position=2,
            split_location=2,
            expected_sources=("2+", "5"),
            cell_type="code",
        )

    def test_split_first_cell(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        uuid = upload_notebook(client, superuser_token_headers, "simple.ipynb")
        self._test_split_cell(
            client,
            superuser_token_headers,
            uuid=uuid,
            position=0,
            split_location=11,
            expected_sources=("## Simple T", "est Notebook"),
            cell_type="markdown",
        )

    def test_split_last_cell(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        uuid = upload_notebook(client, superuser_token_headers, "simple.ipynb")
        self._test_split_cell(
            client,
            superuser_token_headers,
            uuid=uuid,
            position=3,
            split_location=0,
            expected_sources=("", ""),
            cell_type="code",
        )

    def test_round_trip(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        """
        Check that when a notebook is uploaded, it gets a UUID.
        Then when it is download, uploaded and downloaded again,
        the UUID is different.
        """
        # initial upload
        uuid = upload_notebook(client, superuser_token_headers, "simple.ipynb")

        # download the notebook, now with uuid
        response = client.get(
            f"/notebooks/{uuid}", headers=superuser_token_headers
        )
        assert response.status_code == 200
        notebook = response.json()["notebook"]
        data = json.dumps(notebook)
        dl1_uuid = notebook["metadata"]["jupyter_d1"]["uuid"]
        assert uuid == dl1_uuid
        assert (
            notebook["metadata"]["jupyter_d1"]["path"]
            == f"{settings.ROOT_DIR}/simple.ipynb"
        )
        assert notebook["metadata"]["jupyter_d1"]["name"] == "simple"

        # delete the notebook on the server
        response = client.delete(
            f"/notebooks/{uuid}", headers=superuser_token_headers
        )
        assert response.status_code == 204

        # upload the downloaded notebook, fail because file exists
        response = client.post(
            "/notebooks/upload",
            params={"filename": "simple.ipynb"},
            data=data,
            headers=superuser_token_headers,
        )
        assert response.status_code == 400
        path = response.json()["detail"] == "File alreaady exists"

        # upload the downloaded notebook with different name,
        # complete with uuid
        response = client.post(
            "/notebooks/upload",
            params={"filename": "isimple.ipynb"},
            data=data,
            headers=superuser_token_headers,
        )
        assert response.status_code == 201
        path = response.json()["path"]

        # Check that remote file has a different uuid after upload
        with open(f"{settings.ROOT_DIR}/{path}", "r") as f:
            nb = json.loads(f.read())
        file_uuid = nb["metadata"]["jupyter_d1"]["uuid"]
        assert uuid != file_uuid
        assert (
            nb["metadata"]["jupyter_d1"]["path"]
            == f"{settings.ROOT_DIR}/isimple.ipynb"
        )
        assert nb["metadata"]["jupyter_d1"]["name"] == "isimple"

        # Open notebook, check that uuid changes again
        response = client.get(
            f"/notebooks/open/?filepath={path}",
            headers=superuser_token_headers,
        )
        assert response.status_code == 201
        notebook = response.json()["notebook"]
        open_uuid = notebook["metadata"]["jupyter_d1"]["uuid"]
        assert uuid != open_uuid
        assert file_uuid != open_uuid
        assert (
            notebook["metadata"]["jupyter_d1"]["path"]
            == f"{settings.ROOT_DIR}/isimple.ipynb"
        )
        assert notebook["metadata"]["jupyter_d1"]["name"] == "isimple"

        # Check that new uuid was written to disk after opening the notebook
        with open(f"{settings.ROOT_DIR}/{path}", "r") as f:
            nb = json.loads(f.read())
        new_file_uuid = nb["metadata"]["jupyter_d1"]["uuid"]
        assert new_file_uuid == open_uuid
        assert (
            nb["metadata"]["jupyter_d1"]["path"]
            == f"{settings.ROOT_DIR}/isimple.ipynb"
        )
        assert nb["metadata"]["jupyter_d1"]["name"] == "isimple"

        # finally, get the notebook again and make sure the
        # uuid is the same as when opened
        response = client.get(
            f"/notebooks/{open_uuid}", headers=superuser_token_headers
        )
        assert response.status_code == 200
        notebook = response.json()["notebook"]
        dl2_uuid = notebook["metadata"]["jupyter_d1"]["uuid"]
        assert open_uuid == dl2_uuid

    def test_cell_state_reset(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        """
        Check that when a notebook is uploaded, cells states get reset to
        idle and execution counts are reset
        """
        # initial upload
        uuid = upload_notebook(
            client, superuser_token_headers, "stateful_simple.ipynb"
        )

        # download the notebook
        response = client.get(
            f"/notebooks/{uuid}", headers=superuser_token_headers
        )
        assert response.status_code == 200
        notebook = response.json()["notebook"]
        dl1_uuid = notebook["metadata"]["jupyter_d1"]["uuid"]
        assert uuid == dl1_uuid
        assert (
            notebook["metadata"]["jupyter_d1"]["path"]
            == f"{settings.ROOT_DIR}/stateful_simple.ipynb"
        )
        assert notebook["metadata"]["jupyter_d1"]["name"] == "stateful_simple"

        for cell in notebook["cells"]:
            if "jupyter_d1" in cell["metadata"]:
                d1_data = cell["metadata"]["jupyter_d1"]
                if "execution_state" in d1_data:
                    assert d1_data["execution_state"] == "idle"
            if "execution_count" in cell:
                assert cell["execution_count"] is None

    def test_get_bogus_notebook(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        "Test error condition when downloading an non-existent notebook."
        from uuid import uuid4

        uuid = uuid4()
        response = client.get(
            f"/notebooks/{uuid}", headers=superuser_token_headers
        )
        assert response.status_code == 404
        resp_json = response.json()
        assert resp_json["error"] == "NOTEBOOK_NOT_FOUND"
        assert resp_json["reason"] == f"Notebook not found with uuid: {uuid}"
        assert resp_json["detail"] is None

    def test_get_bogus_cell(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        "Test error condition when downloading an non-existent cell."
        uuid = upload_notebook(client, superuser_token_headers, "simple.ipynb")
        from uuid import uuid4

        cell_uuid = uuid4()
        response = client.get(
            f"/notebooks/{uuid}/cells/{cell_uuid}",
            headers=superuser_token_headers,
        )
        assert response.status_code == 404
        resp_json = response.json()
        assert resp_json["error"] == "CELL_NOT_FOUND"
        assert resp_json["reason"] == f"Cell not found with uuid: {cell_uuid}"
        assert resp_json["detail"] is None

    def test_bogus_kernel(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        "Test notebook upload that has an unsupported kernel."
        upload_notebook(client, superuser_token_headers, "weird_kernel.ipynb")

    def test_upload_to_directory(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        os.mkdir(f"{settings.ROOT_DIR}/aproject")
        with open(f"jupyter_d1/tests/notebooks/simple.ipynb", "r") as f:
            nb_raw = f.read()
        response = client.post(
            "/notebooks/upload",
            params={"directory": "aproject", "filename": "koy"},
            data=nb_raw,
            headers=superuser_token_headers,
        )
        assert response.status_code == 201
        path = response.json()["path"]
        assert path == f"aproject/koy.ipynb"

        with open(f"{settings.ROOT_DIR}/{path}", "r") as f:
            nb = json.loads(f.read())
        file_uuid = nb["metadata"]["jupyter_d1"]["uuid"]
        assert file_uuid is not None
        assert (
            nb["metadata"]["jupyter_d1"]["path"]
            == f"{settings.ROOT_DIR}/{path}"
        )
        assert nb["metadata"]["jupyter_d1"]["name"] == "koy"

        response = client.get(
            f"/notebooks/open/?filepath={path}",
            headers=superuser_token_headers,
        )

        assert response.status_code == 201
        nb = response.json()["notebook"]
        uuid = nb["metadata"]["jupyter_d1"]["uuid"]
        assert uuid != file_uuid
        assert (
            nb["metadata"]["jupyter_d1"]["path"]
            == f"{settings.ROOT_DIR}/{path}"
        )
        assert nb["metadata"]["jupyter_d1"]["name"] == "koy"

    def test_upload_to_directory_fail(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        with open(f"jupyter_d1/tests/notebooks/simple.ipynb", "r") as f:
            nb = json.loads(f.read())
        response = client.post(
            "/notebooks/upload",
            params={"directory": "aproject", "filename": "koy"},
            data=nb,
            headers=superuser_token_headers,
        )
        assert response.status_code == 404
        assert (
            response.json()["detail"]
            == f"Directory does not exist {settings.ROOT_DIR}/aproject"
        )

    def test_working_directory(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        nb_filename = f"jupyter_d1/tests/notebooks/simple.ipynb"
        nb_json = open(nb_filename).read()
        response = client.post(
            "/notebooks/upload",
            params={"filename": "name.ipynb"},
            data=nb_json,
            headers=superuser_token_headers,
        )
        assert response.status_code == 201
        assert response.json()["path"] == "name.ipynb"
        response = client.get(
            "/notebooks/open/?filepath=name.ipynb",
            headers=superuser_token_headers,
        )
        assert response.status_code == 201
        resp_json = response.json()
        assert (
            resp_json["notebook"]["metadata"]["jupyter_d1"][
                "working_directory"
            ]
            == settings.ROOT_DIR
        )
        uuid = resp_json["notebook"]["metadata"]["jupyter_d1"]["uuid"]
        response = client.delete(
            f"/notebooks/{uuid}", data=nb_json, headers=superuser_token_headers
        )
        assert response.status_code == 204

        response = client.get(
            "/notebooks/open/?filepath=name.ipynb",
            headers=superuser_token_headers,
            params={"working_directory": "/tmp"},
        )
        assert response.json()["notebook"]["metadata"]["jupyter_d1"][
            "working_directory"
        ] == str(pathlib.Path("/tmp").resolve())

    def test_upload_and_open_in_memory_no_working_dir_specified(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        """
        Working directory should default to root dir, 
        notebook should not be saved on disk.
        """
        nb_filename = f"jupyter_d1/tests/notebooks/simple.ipynb"
        nb_json = open(nb_filename).read()
        response = client.post(
            "/notebooks/upload_and_open",
            params={"filename": "name.ipynb", "autosave": False},
            data=nb_json,
            headers=superuser_token_headers,
        )
        assert response.status_code == 201
        resp_json = response.json()
        assert (
            resp_json["notebook"]["metadata"]["jupyter_d1"][
                "working_directory"
            ]
            == settings.ROOT_DIR
        )
        uuid = resp_json["notebook"]["metadata"]["jupyter_d1"]["uuid"]

        # Do an action that would save the notebook if autosave was true
        cell1_uuid = resp_json["notebook"]["cells"][1]["metadata"][
            "jupyter_d1"
        ]["uuid"]
        new_source = 'hello = "world"'
        response = client.patch(
            f"/notebooks/{uuid}/cells/{cell1_uuid}",
            headers=superuser_token_headers,
            json={"source": new_source},
        )
        assert response.status_code == 200

        assert not os.path.exists(f"{settings.ROOT_DIR}/name.ipynb")

        response = client.get("/notebooks", headers=superuser_token_headers)
        assert response.status_code == 200
        resp_json = response.json()
        notebooks = resp_json["notebooks"]
        assert len(notebooks) == 1
        assert notebooks[0]["metadata"]["jupyter_d1"]["uuid"] == uuid
        assert (
            notebooks[0]["metadata"]["jupyter_d1"]["working_directory"]
            == settings.ROOT_DIR
        )

    def test_upload_and_open_in_memory_file_exists_but_its_fine(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        """
        Opening notebook in memory when file exists should work fine.
        """
        nb_filename = f"jupyter_d1/tests/notebooks/simple.ipynb"
        nb_json = open(nb_filename).read()
        with open(f"{settings.ROOT_DIR}/name.ipynb", "w") as f:
            f.write(nb_json)
        response = client.post(
            "/notebooks/upload_and_open",
            params={"filename": "name.ipynb", "autosave": False},
            data=nb_json,
            headers=superuser_token_headers,
        )
        assert response.status_code == 201
        resp_json = response.json()
        assert (
            resp_json["notebook"]["metadata"]["jupyter_d1"][
                "working_directory"
            ]
            == settings.ROOT_DIR
        )
        uuid = resp_json["notebook"]["metadata"]["jupyter_d1"]["uuid"]

        assert os.path.exists(f"{settings.ROOT_DIR}/name.ipynb")

        response = client.get("/notebooks", headers=superuser_token_headers)
        assert response.status_code == 200
        resp_json = response.json()
        notebooks = resp_json["notebooks"]
        assert len(notebooks) == 1
        assert notebooks[0]["metadata"]["jupyter_d1"]["uuid"] == uuid
        assert (
            notebooks[0]["metadata"]["jupyter_d1"]["working_directory"]
            == settings.ROOT_DIR
        )

    def test_upload_and_open_in_memory_working_dir_specified(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        """
        Working directory should be set to specified directory, 
        notebook should not be saved on disk.
        """
        nb_filename = f"jupyter_d1/tests/notebooks/simple.ipynb"
        nb_json = open(nb_filename).read()
        response = client.post(
            "/notebooks/upload_and_open",
            params={
                "filename": "name.ipynb",
                "autosave": False,
                "working_directory": f"{os.getcwd()}/jupyter_d1"
                "/tests/notebooks",
            },
            data=nb_json,
            headers=superuser_token_headers,
        )
        assert response.status_code == 201
        resp_json = response.json()
        assert (
            resp_json["notebook"]["metadata"]["jupyter_d1"][
                "working_directory"
            ]
            == f"{os.getcwd()}/jupyter_d1/tests/notebooks"
        )
        uuid = resp_json["notebook"]["metadata"]["jupyter_d1"]["uuid"]

        assert not os.path.exists(f"{settings.ROOT_DIR}/name.ipynb")
        # This should be impossible, but might as well check it
        assert not os.path.exists(
            f"{os.getcwd()}/jupyter_d1/tests/notebooks/name.ipynb"
        )

        response = client.get("/notebooks", headers=superuser_token_headers)
        assert response.status_code == 200
        resp_json = response.json()
        notebooks = resp_json["notebooks"]
        assert len(notebooks) == 1
        assert notebooks[0]["metadata"]["jupyter_d1"]["uuid"] == uuid
        assert (
            notebooks[0]["metadata"]["jupyter_d1"]["working_directory"]
            == f"{os.getcwd()}/jupyter_d1/tests/notebooks"
        )

    def test_upload_and_open_no_working_dir_specified(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        """
        Working directory should default to root dir, 
        notebook should be saved on disk.
        """
        nb_filename = f"jupyter_d1/tests/notebooks/simple.ipynb"
        nb_json = open(nb_filename).read()
        response = client.post(
            "/notebooks/upload_and_open",
            params={"filename": "name.ipynb"},
            data=nb_json,
            headers=superuser_token_headers,
        )
        assert response.status_code == 201
        resp_json = response.json()
        assert (
            resp_json["notebook"]["metadata"]["jupyter_d1"][
                "working_directory"
            ]
            == settings.ROOT_DIR
        )
        uuid = resp_json["notebook"]["metadata"]["jupyter_d1"]["uuid"]

        assert os.path.exists(f"{settings.ROOT_DIR}/name.ipynb")

        response = client.get("/notebooks", headers=superuser_token_headers)
        assert response.status_code == 200
        resp_json = response.json()
        notebooks = resp_json["notebooks"]
        assert len(notebooks) == 1
        assert notebooks[0]["metadata"]["jupyter_d1"]["uuid"] == uuid
        assert (
            notebooks[0]["metadata"]["jupyter_d1"]["working_directory"]
            == settings.ROOT_DIR
        )

    def test_upload_and_open_working_dir_defaults_to_root_dir(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        """
        Working directory should default to root dir, 
        notebook should be saved on disk.
        """
        try:
            os.remove(f"{os.getcwd()}/jupyter_d1/tests/notebooks/name.ipynb")
        except Exception:
            pass
        nb_filename = f"jupyter_d1/tests/notebooks/simple.ipynb"
        nb_json = open(nb_filename).read()
        response = client.post(
            "/notebooks/upload_and_open",
            params={
                "filename": "name.ipynb",
                "directory": f"{os.getcwd()}/jupyter_d1/tests/notebooks",
            },
            data=nb_json,
            headers=superuser_token_headers,
        )
        assert response.status_code == 201
        resp_json = response.json()
        assert (
            resp_json["notebook"]["metadata"]["jupyter_d1"][
                "working_directory"
            ]
            == f"{os.getcwd()}/jupyter_d1/tests/notebooks"
        )
        uuid = resp_json["notebook"]["metadata"]["jupyter_d1"]["uuid"]

        assert os.path.exists(
            f"{os.getcwd()}/jupyter_d1/tests/notebooks/name.ipynb"
        )

        response = client.get("/notebooks", headers=superuser_token_headers)
        assert response.status_code == 200
        resp_json = response.json()
        notebooks = resp_json["notebooks"]
        assert len(notebooks) == 1
        assert notebooks[0]["metadata"]["jupyter_d1"]["uuid"] == uuid
        assert (
            notebooks[0]["metadata"]["jupyter_d1"]["working_directory"]
            == f"{os.getcwd()}/jupyter_d1/tests/notebooks"
        )
        os.remove(f"{os.getcwd()}/jupyter_d1/tests/notebooks/name.ipynb")

    def test_upload_and_open_working_dir_defaults_to_dir(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        """
        Working directory should default to directory provided (where notebook
        is also saved), notebook should be saved on disk.
        """
        try:
            os.remove(f"{os.getcwd()}/jupyter_d1/tests/notebooks/name.ipynb")
        except Exception:
            pass
        nb_filename = f"jupyter_d1/tests/notebooks/simple.ipynb"
        nb_json = open(nb_filename).read()
        response = client.post(
            "/notebooks/upload_and_open",
            params={
                "filename": "name.ipynb",
                "directory": f"{os.getcwd()}/jupyter_d1/tests/notebooks",
            },
            data=nb_json,
            headers=superuser_token_headers,
        )
        assert response.status_code == 201
        resp_json = response.json()
        assert (
            resp_json["notebook"]["metadata"]["jupyter_d1"][
                "working_directory"
            ]
            == f"{os.getcwd()}/jupyter_d1/tests/notebooks"
        )
        uuid = resp_json["notebook"]["metadata"]["jupyter_d1"]["uuid"]

        assert os.path.exists(
            f"{os.getcwd()}/jupyter_d1/tests/notebooks/name.ipynb"
        )

        response = client.get("/notebooks", headers=superuser_token_headers)
        assert response.status_code == 200
        resp_json = response.json()
        notebooks = resp_json["notebooks"]
        assert len(notebooks) == 1
        assert notebooks[0]["metadata"]["jupyter_d1"]["uuid"] == uuid
        assert (
            notebooks[0]["metadata"]["jupyter_d1"]["working_directory"]
            == f"{os.getcwd()}/jupyter_d1/tests/notebooks"
        )
        os.remove(f"{os.getcwd()}/jupyter_d1/tests/notebooks/name.ipynb")

    def test_upload_and_open_working_directory_specified(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        """
        Working directory should be set to provided working dir, 
        notebook should be saved on disk.
        """
        try:
            os.remove(f"{os.getcwd()}/jupyter_d1/tests/notebooks/name.ipynb")
        except Exception:
            pass
        nb_filename = f"jupyter_d1/tests/notebooks/simple.ipynb"
        nb_json = open(nb_filename).read()
        response = client.post(
            "/notebooks/upload_and_open",
            params={
                "filename": "name.ipynb",
                "directory": f"{os.getcwd()}/jupyter_d1/tests/notebooks",
                "working_directory": "/tmp",
            },
            data=nb_json,
            headers=superuser_token_headers,
        )
        assert response.status_code == 201
        resp_json = response.json()
        assert resp_json["notebook"]["metadata"]["jupyter_d1"][
            "working_directory"
        ] == str(pathlib.Path("/tmp").resolve())
        uuid = resp_json["notebook"]["metadata"]["jupyter_d1"]["uuid"]

        assert os.path.exists(
            f"{os.getcwd()}/jupyter_d1/tests/notebooks/name.ipynb"
        )

        response = client.get("/notebooks", headers=superuser_token_headers)
        assert response.status_code == 200
        resp_json = response.json()
        notebooks = resp_json["notebooks"]
        assert len(notebooks) == 1
        assert notebooks[0]["metadata"]["jupyter_d1"]["uuid"] == uuid
        assert notebooks[0]["metadata"]["jupyter_d1"][
            "working_directory"
        ] == str(pathlib.Path("/tmp").resolve())
        os.remove(f"{os.getcwd()}/jupyter_d1/tests/notebooks/name.ipynb")

    def test_open_notebook_file_twice_returns_same_notebook(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        """Opening a notebook twice returns an error"""
        nb_filename = f"jupyter_d1/tests/notebooks/simple.ipynb"
        nb_json = open(nb_filename).read()
        response = client.post(
            "/notebooks/upload",
            params={"filename": "name.ipynb"},
            data=nb_json,
            headers=superuser_token_headers,
        )
        assert response.status_code == 201
        assert response.json()["path"] == "name.ipynb"
        response = client.get(
            "/notebooks/open/?filepath=name.ipynb",
            headers=superuser_token_headers,
        )
        assert response.status_code == 201
        metadata = response.json()["notebook"]["metadata"]["jupyter_d1"]
        assert metadata["working_directory"] == settings.ROOT_DIR
        uuid = metadata["uuid"]

        response = client.get(
            "/notebooks/open/?filepath=name.ipynb",
            headers=superuser_token_headers,
            params={"working_directory": "/tmp"},
        )
        assert response.status_code == 201
        metadata = response.json()["notebook"]["metadata"]["jupyter_d1"]
        assert metadata["working_directory"] == settings.ROOT_DIR
        assert metadata["uuid"] == uuid

        response = client.get("/notebooks", headers=superuser_token_headers)
        assert response.status_code == 200
        resp_json = response.json()
        notebooks = resp_json["notebooks"]
        assert len(notebooks) == 1
        assert notebooks[0]["metadata"]["jupyter_d1"]["uuid"] == uuid
        assert (
            notebooks[0]["metadata"]["jupyter_d1"]["working_directory"]
            == settings.ROOT_DIR
        )

    def test_upload_and_open_notebook_twice_returns_same_notebook(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        nb_filename = f"jupyter_d1/tests/notebooks/simple.ipynb"
        nb_json = open(nb_filename).read()
        response = client.post(
            "/notebooks/upload_and_open",
            params={"autosave": False},
            data=nb_json,
            headers=superuser_token_headers,
        )
        assert response.status_code == 201
        resp_json = response.json()
        assert (
            resp_json["notebook"]["metadata"]["jupyter_d1"][
                "working_directory"
            ]
            == settings.ROOT_DIR
        )
        uuid = resp_json["notebook"]["metadata"]["jupyter_d1"]["uuid"]

        assert not os.path.exists(f"{settings.ROOT_DIR}/name.ipynb")

        response = client.post(
            "/notebooks/upload_and_open",
            params={"autosave": False, "working_directory": "/tmp"},
            data=json.dumps(resp_json["notebook"]),
            headers=superuser_token_headers,
        )
        assert response.status_code == 201
        resp_json = response.json()
        assert (
            resp_json["notebook"]["metadata"]["jupyter_d1"][
                "working_directory"
            ]
            == settings.ROOT_DIR
        )
        assert resp_json["notebook"]["metadata"]["jupyter_d1"]["uuid"] == uuid

        response = client.get("/notebooks", headers=superuser_token_headers)
        assert response.status_code == 200
        resp_json = response.json()
        notebooks = resp_json["notebooks"]
        assert len(notebooks) == 1
        assert notebooks[0]["metadata"]["jupyter_d1"]["uuid"] == uuid
        assert (
            notebooks[0]["metadata"]["jupyter_d1"]["working_directory"]
            == settings.ROOT_DIR
        )

    def test_upload_invalid_notebook(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        """
        Invalid notebook should return 400, and not cause a 500
        """
        response = client.post(
            "/notebooks/upload",
            params={"filename": "name.ipynb", "autosave": False},
            data="{}",
            headers=superuser_token_headers,
        )
        assert response.status_code == 400
        assert response.json()["detail"] == \
            "Failed to parse notebook: <class 'AttributeError'> cells"

    def test_open_invalid_notebook(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        """
        Invalid notebook should return 400, and not cause a 500
        """
        with open(f"{settings.ROOT_DIR}/name1l.ipynb", "w") as f:
            f.write("{}")

        response = client.get(
            "/notebooks/open/",
            params={"filepath": "name1l.ipynb"},
            data="{}",
            headers=superuser_token_headers,
        )
        assert response.status_code == 400
        assert response.json()["detail"] == \
            "Failed to parse notebook: <class 'AttributeError'> cells"

    def test_upload_and_open_invalid_notebook(
        self, client: TestClient, superuser_token_headers: Dict[str, str]
    ):
        """
        Invalid notebook should return 400, and not cause a 500
        """
        response = client.post(
            "/notebooks/upload_and_open",
            params={"filename": "name.ipynb", "autosave": False},
            data="{}",
            headers=superuser_token_headers,
        )
        assert response.status_code == 400
        assert response.json()["detail"] == \
            "Failed to parse notebook: <class 'AttributeError'> cells"


@pytest.mark.usefixtures("clear_notebooks", "clear_notebook_directory")
class TestNotebookPermissions:
    def test_notebooks(
        self,
        client: TestClient,
        superuser_token_headers: Dict[str, str],
        readonly_token_headers: Dict[str, str],
        permissionless_token_headers: Dict[str, str],
    ):
        upload_notebook(client, superuser_token_headers, "simple.ipynb")
        upload_notebook(client, superuser_token_headers, "other_simple.ipynb")
        response = client.get(
            "/notebooks", headers=permissionless_token_headers
        )
        assert response.status_code == 403
        response = client.get("/notebooks", headers=readonly_token_headers)
        assert response.status_code == 200

    def test_notebook(
        self,
        client: TestClient,
        readonly_token_headers: Dict[str, str],
        permissionless_token_headers: Dict[str, str],
        superuser_token_headers: Dict[str, str],
    ):
        uuid = upload_notebook(client, superuser_token_headers)
        response = client.get(
            f"/notebooks/{uuid}", headers=permissionless_token_headers
        )
        assert response.status_code == 403
        response = client.get(
            f"/notebooks/{uuid}", headers=readonly_token_headers
        )
        assert response.status_code == 200

    def test_delete_notebooks(
        self,
        client: TestClient,
        readonly_token_headers: Dict[str, str],
        superuser_token_headers: Dict[str, str],
    ):
        upload_notebook(client, superuser_token_headers, "simple.ipynb")
        upload_notebook(client, superuser_token_headers, "other_simple.ipynb")
        response = client.delete("/notebooks", headers=readonly_token_headers)
        assert response.status_code == 403
        response = client.delete("/notebooks", headers=superuser_token_headers)
        assert response.status_code == 204

    def test_delete_notebook(
        self,
        client: TestClient,
        readonly_token_headers: Dict[str, str],
        superuser_token_headers: Dict[str, str],
    ):
        uuid = upload_notebook(client, superuser_token_headers, "simple.ipynb")
        response = client.delete(
            f"/notebooks/{uuid}", headers=readonly_token_headers
        )
        assert response.status_code == 403
        response = client.delete(
            f"/notebooks/{uuid}", headers=superuser_token_headers
        )
        assert response.status_code == 204

    def test_cells(
        self,
        client: TestClient,
        readonly_token_headers: Dict[str, str],
        permissionless_token_headers: Dict[str, str],
        superuser_token_headers: Dict[str, str],
    ):
        uuid = upload_notebook(client, superuser_token_headers, "simple.ipynb")

        response = client.get(
            f"/notebooks/{uuid}/cells", headers=permissionless_token_headers
        )
        assert response.status_code == 403
        response = client.get(
            f"/notebooks/{uuid}/cells", headers=readonly_token_headers
        )
        assert response.status_code == 200

    def test_one_cell(
        self,
        client: TestClient,
        readonly_token_headers: Dict[str, str],
        permissionless_token_headers: Dict[str, str],
        superuser_token_headers: Dict[str, str],
    ):
        uuid = upload_notebook(client, superuser_token_headers, "simple.ipynb")

        response = client.get(
            f"/notebooks/{uuid}/cells", headers=readonly_token_headers
        )
        assert response.status_code == 200
        cell1_uuid = response.json()["cells"][1]["metadata"]["jupyter_d1"][
            "uuid"
        ]

        response = client.get(
            f"/notebooks/{uuid}/cells/{cell1_uuid}",
            headers=permissionless_token_headers,
        )
        assert response.status_code == 403
        response = client.get(
            f"/notebooks/{uuid}/cells/{cell1_uuid}",
            headers=readonly_token_headers,
        )
        assert response.status_code == 200

    def test_patch_cell(
        self,
        client: TestClient,
        readonly_token_headers: Dict[str, str],
        superuser_token_headers: Dict[str, str],
    ):
        uuid = upload_notebook(client, superuser_token_headers, "simple.ipynb")

        response = client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cell1_uuid = response.json()["cells"][1]["metadata"]["jupyter_d1"][
            "uuid"
        ]

        new_source = 'hello = "world"'
        response = client.patch(
            f"/notebooks/{uuid}/cells/{cell1_uuid}",
            headers=readonly_token_headers,
            json={"source": new_source},
        )
        assert response.status_code == 403
        response = client.patch(
            f"/notebooks/{uuid}/cells/{cell1_uuid}",
            headers=superuser_token_headers,
            json={"source": new_source},
        )
        assert response.status_code == 200

    def test_patch_and_execute_cell(
        self,
        client: TestClient,
        readonly_token_headers: Dict[str, str],
        superuser_token_headers: Dict[str, str],
    ):
        uuid = upload_notebook(client, superuser_token_headers, "simple.ipynb")

        response = client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        cell1_uuid = response.json()["cells"][1]["metadata"]["jupyter_d1"][
            "uuid"
        ]

        new_source = 'hello = "world"'
        response = client.patch(
            f"/notebooks/{uuid}/cells/{cell1_uuid}/execute",
            headers=readonly_token_headers,
            json={"source": new_source},
        )
        assert response.status_code == 403
        response = client.patch(
            f"/notebooks/{uuid}/cells/{cell1_uuid}/execute",
            headers=superuser_token_headers,
            json={"source": new_source},
        )
        assert response.status_code == 200

        uuid = response.json()["cell"]["metadata"]["jupyter_d1"]["uuid"]
        assert len(uuid) == 36
        message_id = response.json()["kernel_message"]["message_id"]
        assert len(message_id) == 35

    def test_create_cell(
        self,
        client: TestClient,
        readonly_token_headers: Dict[str, str],
        superuser_token_headers: Dict[str, str],
    ):
        uuid = upload_notebook(client, superuser_token_headers, "simple.ipynb")

        # add a new cell
        src = "__**Hello Callist**__"
        params = {"cell_type": "markdown", "source": src}
        response = client.post(
            f"/notebooks/{uuid}/cells",
            json=params,
            headers=readonly_token_headers,
        )
        assert response.status_code == 403
        response = client.post(
            f"/notebooks/{uuid}/cells",
            json=params,
            headers=superuser_token_headers,
        )
        assert response.status_code == 201

    def test_move_cell(
        self,
        client: TestClient,
        readonly_token_headers: Dict[str, str],
        superuser_token_headers: Dict[str, str],
    ):
        uuid = upload_notebook(client, superuser_token_headers, "simple.ipynb")

        # get the existing cells
        response = client.get(
            f"/notebooks/{uuid}/cells", headers=superuser_token_headers
        )
        assert response.status_code == 200
        move_uuid = response.json()["cells"][2]["metadata"]["jupyter_d1"][
            "uuid"
        ]

        response = client.get(
            f"/notebooks/{uuid}/cells/{move_uuid}/move",
            headers=readonly_token_headers,
        )
        assert response.status_code == 403
        response = client.get(
            f"/notebooks/{uuid}/cells/{move_uuid}/move",
            headers=superuser_token_headers,
        )
        assert response.status_code == 204

    def test_upload_notebook(
        self,
        client: TestClient,
        readonly_token_headers: Dict[str, str],
        superuser_token_headers: Dict[str, str],
    ):
        nb_filename = f"jupyter_d1/tests/notebooks/simple.ipynb"
        nb_json = open(nb_filename).read()
        response = client.post(
            "/notebooks/upload", data=nb_json, headers=readonly_token_headers
        )
        assert response.status_code == 403
        response = client.post(
            "/notebooks/upload",
            params={"filename": "name.ipynb"},
            data=nb_json,
            headers=superuser_token_headers,
        )
        assert response.status_code == 201
        assert response.json()["path"] == "name.ipynb"
        response = client.get(
            "/notebooks/open/?filepath=name.ipynb",
            headers=superuser_token_headers,
        )
        assert response.status_code == 201



from abc import ABC, abstractmethod
from typing import List, Optional

from ..models.kernel_variable import KernelVariable
from ..utils import NotebookNode

# TODO: Send updates as diffs


class VarsManager(ABC):
    def __init__(self):
        self._vars = []
        self._tmp_vars = []

    @abstractmethod
    def get_vars_code(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def get_single_var_code(self, var_name: str) -> str:
        raise NotImplementedError

    @abstractmethod
    def parse_vars_response(
        self, vars_response: NotebookNode
    ) -> List[KernelVariable]:
        raise NotImplementedError

    @abstractmethod
    def parse_single_var_response(
        self, var_response: NotebookNode
    ) -> Optional[KernelVariable]:
        return None

    @property
    def vars(self) -> List[KernelVariable]:
        return self._vars

    def parse_output(self, vars_output: NotebookNode):
        vars = self.parse_vars_response(vars_output)
        self._tmp_vars += vars

    def on_request_start(self):
        self._tmp_vars = []

    def on_request_end(self) -> List[KernelVariable]:
        self._vars = self._tmp_vars
        return self._vars



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="callisto-jupyter-d1",
    version="0.0.5",
    author="Oak City Labs",
    author_email="team@oakcity.io",
    description="Jupyter D1 Server for Callisto",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://callistoapp.com",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        "wsgidav>=3.0.3",
        "fastapi>=0.52.0",
        "asyncblink>=0.3.2",
        "uvicorn[standard]>=0.11.3",
        "python-dotenv>=0.13.0",
        "nbformat>=5.0.4",
        "python-jose>=3.1.0",
        "watchdog>=2.1.7",
        "jupyter>=1.0.0",
        "jupyter_client>=6.1.6",
        "jupyter_console>=6.1.0",
        "jupyter_core>=4.6.3",
        "jupyter_kernel_gateway>=2.4.0",
        "zsh-jupyter-kernel>=3.2",
        "bash_kernel>=0.7.2",
        "pynvml==8.0.4",
        "callisto-python==0.0.2",
        "jupytext==1.13.8",
        "python-multipart>=0.0.5",
    ],
    extras_require={
        "full": ["psutil==5.7.2"]
    },
    scripts=["start_jupyter_d1", "jupyter_d1_test",
             "jupyter_d1_install_kernels"]
)



from datetime import datetime, timedelta
from typing import Dict

import pytest  # type: ignore
from async_asgi_testclient import (
    TestClient as WebSocketTestClient,  # type: ignore
)
from fastapi.testclient import TestClient
from pytest_mock import MockFixture  # type: ignore

from .utils import wait_for_event


def test_rclone(
    client: TestClient,
    superuser_token_headers: Dict[str, str],
    mocker: MockFixture,
):
    mock_requests = mocker.patch("jupyter_d1.routers.rclone.requests")
    mock_requests.post().json.return_value = {"pldefe": "fj39499"}
    response = client.post(
        "/rclone/config/dump", headers=superuser_token_headers
    )
    assert response.status_code == 200
    # print(response.json())
    assert response.json() == {"pldefe": "fj39499"}
    assert (
        mock_requests.post.call_args[0][0]
        == "http://127.0.0.1:5572/config/dump"
    )


def test_rclone_query_params(
    client: TestClient,
    superuser_token_headers: Dict[str, str],
    mocker: MockFixture,
):
    mock_requests = mocker.patch("jupyter_d1.routers.rclone.requests")
    mock_requests.post().json.return_value = {"gg": "754d"}
    response = client.post(
        "/rclone/config/get",
        headers=superuser_token_headers,
        params={"name": "gdrive"},
    )
    assert response.status_code == 200
    # print(response.json())
    assert response.json() == {"gg": "754d"}
    assert (
        mock_requests.post.call_args[0][0]
        == "http://127.0.0.1:5572/config/get"
    )
    assert dict(mock_requests.post.call_args[1]["params"]) == {
        "name": "gdrive"
    }


def test_rclone_post_data(
    client: TestClient,
    superuser_token_headers: Dict[str, str],
    mocker: MockFixture,
):
    mock_requests = mocker.patch("jupyter_d1.routers.rclone.requests")
    mock_requests.post().json.return_value = {"polnd": "logkj"}
    superuser_token_headers["Content-Type"] = "application/json"
    response = client.post(
        "/rclone/options/set",
        headers=superuser_token_headers,
        data='{"vfs": {"CacheMaxSize": 100}}',
    )
    assert response.status_code == 200
    # print(response.json())
    assert response.json() == {"polnd": "logkj"}
    # mock_requests.post.assert_called_once()
    # print(mock_requests.post.calls)
    assert (
        mock_requests.post.call_args[0][0]
        == "http://127.0.0.1:5572/options/set"
    )
    assert (
        mock_requests.post.call_args[1]["data"]
        == b'{"vfs": {"CacheMaxSize": 100}}'
    )


@pytest.fixture()
async def superuser_token_headers(
    websocket_client: WebSocketTestClient,
) -> Dict[str, str]:
    r = await websocket_client.get(
        "/login/access-token", headers={"Authorization": "test9token_4"}
    )
    token = r.json()["token"]["access_token"]
    headers = {"Authorization": f"Bearer {token}"}
    return headers


@pytest.mark.asyncio
async def test_rclone_websocket(
    websocket_client: WebSocketTestClient,
    superuser_token_headers: Dict[str, str],
    mocker: MockFixture,
):
    mock_requests = mocker.patch("jupyter_d1.routers.rclone.requests")
    mock_requests.post().json.return_value = {"polnd": "logkj"}

    async with websocket_client.websocket_connect(
        f"/server/ws", headers=superuser_token_headers
    ) as websocket:
        time = datetime.now()

        async def assert_stats():
            resp = await wait_for_event(websocket, "rclone_stats")
            assert resp == {"rclone_stats": {"polnd": "logkj"}}
            assert (
                time - datetime.now() - timedelta(seconds=3)
            ).microseconds < 1e6

        for i in range(3):
            await assert_stats()
